# -*- coding: utf-8 -*-
#"""
#Created on Sun Jun 22 20:00:00 2014
#
#@author: aaron
#"""

import sys
import ioFunctions as io            ### Module for write on disk functions.
import config as cfg                ### Module for global variables for Quantum Walk.
import parsing as par               ### Module for parsing functions for the input file and setting variables
import time                         ### Library for the runtime.
import dtqw1d
import dtqw2d
import staggered1d
import staggered2d
import custom
import os
import distance as dist


### Checking existing file or TEST_MODE

def walk(inputFile):

    ### Acquiring the initial time
    t0=time.time()    
    
#    ###PreParsing
    par.preParsing(inputFile)
    
    ### Parsing the input file
    par.parsingFile(inputFile)

    ### Setting work path
    io.directory_Creation()
    
    ### Generating some files for Quantum Walk
    io.initializeFiles()
    
    if cfg.ADJMATRIX_PATH != "@NON_INICIALIZED@":
        dist.generateDistance()
        
        
    if cfg.WALK=="DTQW1D":
        cfg.RETURN_SIMULATION_FLAG=dtqw1d.run()
    
    elif cfg.WALK=="DTQW2D":
        cfg.RETURN_SIMULATION_FLAG=dtqw2d.run()
    
    elif cfg.WALK=="STAGGERED1D":
        cfg.RETURN_SIMULATION_FLAG=staggered1d.run()

    elif cfg.WALK=="STAGGERED2D":
        cfg.RETURN_SIMULATION_FLAG=staggered2d.run()
    
    elif cfg.WALK=="SZEGEDY":
        cfg.RETURN_SIMULATION_FLAG=szegedy.run()    
    
    elif cfg.WALK=="CUSTOM":
        cfg.RETURN_SIMULATION_FLAG=custom.run()
         
    ### Cleaning temporary files
    if not cfg.DEBUG:
        io.cleaningTemporaryFiles()

    
    os.chdir(cfg.OLD_DIRECTORY)
    
    
    
    ### End runtime
    t1=time.time()

    print("[HIPERWALK] Runtime: ",t1-t0)

    return cfg.RETURN_SIMULATION_FLAG
    



import numpy as np

def _binary_search(v, elem, start=0, end=None):
    r"""
    This function expects a sorted array and performs a binary search on the subarray 
    v[start:end], looking for the element 'elem'. 
    If the element is found, the function returns the index of the element. 
    If the element is not found, the function returns -1.    
    This is an implementation of binary search following Cormen's algorithm. 
    It is used to improve the time complexity of the search process.
    """

    if end == None:
        end = len(v)
    
    while start < end:
        mid = int((start + end)/2)
        if elem <= v[mid]:
            end = mid
        else:
            start = mid + 1

    return end if v[end] == elem else -1

class Graph():
    r"""
    Arbitrary graph.

    The graph on which a quantum walk takes place.

    Parameters
    ----------
    adj_matrix : :class:`scipy.sparse.csr_array`
        Adjacency matrix of the graph on
        which the quantum walk takes place.

    Raises
    ------
    TypeError
        if ``adj_matrix`` is not an instance of
        :class:`scipy.sparse.csr_array`.

    Notes
    -----
    A wide range of methods are available. 
    These methods can be used by a quantum walk model 
    to generate a valid quantum walk.
    
    This class can be passed as an argument to plotting functions.
    In this case,
    the default representation for the graph will be displayed.

    The recommended parameter type is
    :class:`scipy.sparse.csr_array` with ``dtype=np.int8``,
    where 1 denotes adjacency and 0 denotes non-adjacency.
    If any entry differs from 0 or 1,
    some methods may not operate as expected.

    Each edge of a given graph :math:`G(V, E)` is associated with
    two arcs in the symmetric digraph :math:`\vec{G}(V, A)`,
    where

    .. math::
        \begin{align*}
            A = \bigcup_{(v,u) \in E} \{(v, u), (u, v)\}.
        \end{align*}

    An arc can be described either in the arc notation as (tail,head) 
    or through a numerical label. The ordering of the labels in 
    the :obj:`Graph` class is as follows: 
    Consider two arcs, :math:`(v_1, u_1)` and :math:`(v_2, u_2')`, 
    with numerical labels :math:`a_1` and :math:`a_2` respectively. 
    Then, :math:`a_1 < a_2` if and only if either :math:`v_1 < v_2` 
    or :math:`v_1 = v_2` and :math:`u_1 < u_2`.

    .. note::
        The arc ordering may change for graphs defined using specific classes.

    For example, the graph :math:`G(V, E)` shown in
    Figure 1 has an adjacency matrix ``adj_matrix``.

    .. testsetup::

        import numpy as np

    >>> adj_matrix = np.array([
    ...     [0, 1, 0, 0],
    ...     [1, 0, 1, 1],
    ...     [0, 1, 0, 1],
    ...     [0, 1, 1, 0]])
    >>> adj_matrix
    array([[0, 1, 0, 0],
           [1, 0, 1, 1],
           [0, 1, 0, 1],
           [0, 1, 1, 0]])

    .. graphviz:: ../../graphviz/graph-example.dot
        :align: center
        :layout: neato
        :caption: Figure 1

    The arcs of the associated digraph in the arc notation are

    >>> arcs = [(i, j) for i in range(4)
    ...                for j in range(4) if adj_matrix[i,j] == 1]
    >>> arcs
    [(0, 1), (1, 0), (1, 2), (1, 3), (2, 1), (2, 3), (3, 1), (3, 2)]

    Note that ``arcs`` is already sorted, hence the associated 
    numeric labels are

    >>> arcs_labels = {arcs[i]: i for i in range(len(arcs))}
    >>> arcs_labels
    {(0, 1): 0, (1, 0): 1, (1, 2): 2, (1, 3): 3, (2, 1): 4, (2, 3): 5, (3, 1): 6, (3, 2): 7}

    The numeric labels are depicted in Figure 2.

    .. graphviz:: ../../graphviz/graph-arcs.dot
        :align: center
        :layout: neato
        :caption: Figure 2

    If we insert the labels of the arcs into the adjacency matrix,
    we obtain matrix ``adj_labels`` as follows:

    >>> adj_labels = [[arcs_labels[(i,j)] if (i,j) in arcs_labels
    ...                                   else '' for j in range(4)]
    ...               for i in range(4)]
    >>> adj_labels = np.matrix(adj_labels)
    >>> adj_labels
    matrix([['', '0', '', ''],
            ['1', '', '2', '3'],
            ['', '4', '', '5'],
            ['', '6', '7', '']], dtype='<U21')

    Note that, intuitively,
    the arcs are labeled in left-to-right and top-to-bottom fashion.
    """

    def __init__(self, adj_matrix):
        # TODO:
        # * Check if valid adjacency matrix
        # * Add option: numpy dense matrix as parameters.
        # * Add option: networkx graph as parameter.
        self.adj_matrix = adj_matrix
        self.coloring = None

    def default_coin(self):
        r"""
        Returns the default coin for the given graph.

        The default coin for a coined quantum walk on an arbitrary
        graph is ``grover``.
        """
        return 'grover'

    def embeddable(self):
        r"""
        Returns ``True`` if the graph can be embedded on the plane,
        and ``False`` otherwise.

        If a graph can be embedded on the plane,
        we can assign meaningful right or left directions,
        and clockwise or counter-clockwise rotations to edges and arcs

        Notes
        -----
        The implementation is class-dependent.
        We do not inspect the structure of the graph to determine whether
        it is embeddable or not.
        """
        return False

    def arc_label(self, tail, head):
        r"""
        Returns the numerical label of the arc.

        Parameters
        ----------
        tail: int
            Tail of the arc.

        head: int
            Head of the arc.

        Returns
        -------
        label: int
            Arc label.
        """
        return _binary_search(self.adj_matrix.indices, head,
                              start = self.adj_matrix.indptr[tail],
                              end = self.adj_matrix.indptr[tail + 1])

    def arc(self, label):
        r"""
        Converts the numerical label to arc notation.

        Given the arc label as a number,
        returns the arc in the ``(tail, head)`` notation.

        Parameters
        ----------
        label: int
            Arc label (number)

        Returns
        -------
        (int, int)
            Arc in the arc notation ``(tail, head)``.
        """
        adj_matrix = self.adj_matrix
        head = adj_matrix.indices[label]
        # TODO: binary search
        for tail in range(len(adj_matrix.indptr)):
            if adj_matrix.indptr[tail + 1] > label:
                break
        return (tail, head)

    def next_arc(self, arc):
        r"""
        Next arc in an embeddable graph.

        Parameters
        ----------
        arc
            The arc in any of the following notations.

            * arc notation: tuple of vertices
                In ``(tail, head)`` format where
                ``tail`` and ``head`` must be valid vertices.
            * arc label: int.
                The arc label (number).

        Returns
        -------
        Next arc in the same notation as the ``arc`` argument.

        See Also
        --------
        arc
        arc_label
        """
        # implemented only if is embeddable
        raise AttributeError

    def previous_arc(self, arc):
        r"""
        Previous arc in an embeddable graph.

        Parameters
        ----------
        arc
            The arc in any of the following notations.

            * arc notation: tuple of vertices
                In ``(tail, head)`` format where
                ``tail`` and ``head`` must be valid vertices.
            * arc label: int.
                The arc label (number).

        Returns
        -------
        Previous arc in the same notation as the ``arc`` argument.

        See Also
        --------
        arc
        arc_label
        """
        # implemented only if is embeddable
        raise AttributeError

    def neighbors(self, vertex):
        r"""
        Returns all neighbors of the given vertex.
        """
        start = self.adj_matrix.indptr[vertex]
        end = self.adj_matrix.indptr[vertex + 1]
        return self.adj_matrix.indices[start:end]

    def arcs_with_tail(self, tail):
        r"""
        Returns all arcs that have the given tail.
        """
        arcs_lim = self.adj_matrix.indptr
        return np.arange(arcs_lim[tail], arcs_lim[tail + 1])

    def number_of_vertices(self):
        r"""
        Cardinality of vertex set.
        """
        return self.adj_matrix.shape[0]

    def number_of_arcs(self):
        r"""
        Cardinality of arc set.

        For simple graphs, the cardinality is twice the number of edges.
        """
        return self.adj_matrix.sum()

    def number_of_edges(self):
        r"""
        Cardinality of edge set.
        """
        return self.adj_matrix.sum() >> 1

    def degree(self, vertex):
        r"""
        Degree of given vertex.
        """
        indptr = self.adj_matrix.indptr
        return indptr[vertex + 1] - indptr[vertex]



import numpy as np
from sys import path as sys_path
sys_path.append('../')
sys_path.append('../../')
from test_constants import *
import hiperwalk as hpw
import unittest

class TestHypercube(unittest.TestCase):
    
    def setUp(self):
        self.dim = 10
        self.g = hpw.Hypercube(self.dim)

    def tearDown(self):
        del self.g

    def test_arc_direction(self):
        array = [self.g.arc_direction(a)
                 for a in range(self.g.number_of_arcs())]
        num_vert = self.g.number_of_vertices()
        self.assertTrue(array == list(range(self.dim))*num_vert)

    def test_arc_label(self):
        array = [self.g.arc_label(v, v ^ 1 << direction)
                 for v in range(self.g.number_of_vertices())
                 for direction in range(self.dim)]
        self.assertTrue(array == list(range(self.g.number_of_arcs())))

    def test_arc(self):
        array = [self.g.arc_label(self.g.arc(a)[0], self.g.arc(a)[1])
                 for a in range(self.g.number_of_arcs())]
        self.assertTrue(array == list(range(self.g.number_of_arcs())))



print('digraph {')

dim = 3

def out_of_bounds(axis):
    return axis < 0 or axis >= dim 

for y in range(-1, dim + 1):
    for x in range(-1, dim + 1):
        out_x = out_of_bounds(x)
        out_y = out_of_bounds(y)
        if out_x and out_y:
            continue

        node_str = ('\t"' + str((x, y)) + '" ['
                    + 'pos="' + str(1.75*x) + ',' + str(1.75*y) + '!" '
                    + 'width=0.75 height=0.75 fixedsize=True')
        if out_x or out_y:
            node_str += (' style="dashed" label="' +
                         str((x % dim, y % dim)) + '"')
        node_str += ']'

        print(node_str)

print()

arc_count = 0
for y in range(dim):
    for x in range(dim):
        tail = str((x, y))
        for d in range(4):
            y_axis = d // 2
            shift = 1 if d % 2 == 0 else -1
            head = (str((x + shift, y)) if not y_axis else
                    str((x, y + shift)))

            arc_str = '\t "' + tail + '" -> "' + head + '"'
            arc_str += '[headlabel=' + str(arc_count)
            arc_str += ' labeldistance=4'
            arc_str += ' labelangle=-20'
            arc_str += '];'

            print(arc_str)
            arc_count += 1

print('}')



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.append(os.path.abspath('../'))


# -- Project information -----------------------------------------------------

project = 'Hiperwalk'
copyright = '2023'
author = 'Gustavo Bezerra'

# The full version, including alpha/beta/rc tags
release = 'stable'
version = 'stable'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autosummary',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.todo', #'sphinx_autodoc_typehints',
    'numpydoc',
    'sphinx.ext.graphviz',
    'sphinx.ext.doctest',
]

intersphinx_mapping = {
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'networkx' : ('https://networkx.org/documentation/stable/', None),
}

autosummary_generate = True
autoclass_content = "both"
set_type_checking_flag = True
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# Removes numpy auto documentation.
# Documentation is generated recursively according to _templates files
numpydoc_show_class_members = False
#numpydoc_class_members_toctree = False

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'

html_theme_options = {
    "logo" : {
        "text": "Hiperwalk",
        "alt_text": "Hiperwalk",
    },
    "switcher": {
        "json_url": "https://hiperwalk.org/switcher.json",
        "version_match": version,
    },
   "navbar_start": ["navbar-logo", "version-switcher"]
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

master_doc = 'index'



import numpy as np
import scipy.sparse
import scipy.linalg
from .quantum_walk import QuantumWalk
from .._constants import PYNEBLINA_IMPORT_ERROR_MSG
try:
    from . import _pyneblina_interface as nbl
except:
    pass

class ContinuousWalk(QuantumWalk):
    r"""
    Manage an instance of a continuous-time quantum walk
    on any simple graph.

    For further implementation details, refer to the Notes Section.

    Parameters
    ----------
    graph :
        Graph on which the quantum walk takes place.
        There are two acceptable inputs:

        :class:`hiperwalk.graph.Graph` :
            The graph itself.

        :class:`class:scipy.sparse.csr_array`:
            The adjacency matrix of the graph.

    adjacency : :class:`scipy.sparse.csr_array`, optional
        .. deprecated:: 2.0a1
            It will be removed in version 2.0.
            Use ``graph`` instead.

        Adjacency matrix of the graph on which
        the quantum walk takes place.

    **kwargs : optional
        Arguments to set the Hamiltonian.

    See Also
    --------
    set_hamiltonian

    Notes
    -----
    The adjacency matrix of a graph :math:`G(V, E)` is
    the :math:`|V| \times |V|`-dimensional matrix :math:`A` such that
    
    .. math::
        A_{i,j} = \begin{cases}
            1, \text{ if } (i,j) \in E(G),\\
            0, \text{ otherwise.}
        \end{cases}

    The Hamiltonian, which depends on the adjacency matrix and the location of 
    the marked vertices, is described in the
    :meth:`hiperwalk.ContinuousWalk.set_hamiltonian` method.

    The states of the computational basis are :math:`\ket{i}` for
    :math:`0 \leq i < |V|`, where
    :math:`\ket i` is associated with the :math:`i`-th vertex.

    This class can also facilitate the simulation of any Hamiltonian
    evolution. To do this, simply pass the desired Hamiltonian in place
    of the adjacency matrix.
    """

    _hamiltonian_kwargs = dict()

    def __init__(self, graph=None, adjacency=None, **kwargs):

        super().__init__(graph=graph, adjacency=adjacency)

        self.hilb_dim = self._graph.number_of_vertices()
        self._hamiltonian = None

        import inspect

        if not bool(ContinuousWalk._hamiltonian_kwargs):
            # assign static attribute
            ContinuousWalk._hamiltonian_kwargs = {
                'gamma': ContinuousWalk._get_valid_kwargs(self.set_gamma),
                'marked': ContinuousWalk._get_valid_kwargs(self.set_marked)
            }

        self.set_hamiltonian(**kwargs)

    def set_gamma(self, gamma=None):
        r"""
        Sets the gamma parameter.
        
        The gamma parameter is used in the definition of the Hamiltonian.

        Parameters
        ----------
        gamma : float, default = 1
            Gamma value.
        """
        if gamma is None or gamma.imag != 0:
            raise TypeError("Value of 'gamma' is not float.")

        self._gamma = gamma
        self._hamiltonian = None
        self._evolution = None

    def set_marked(self, marked=[]):
        super().set_marked(marked)
        self._hamiltonian = None

    def get_gamma(self):
        r"""
        Retrieves the gamma value used in
        the definition of the Hamiltonian.

        Returns
        -------
        float
        """
        return self._gamma

    def set_hamiltonian(self, **kwargs):
        r"""
        Creates the Hamiltonian.


        Parameters
        ----------
        **kwargs :
            Additional arguments.
            Used for determining the gamma value and marked vertices.
            See :meth:`hiperwalk.ContinuousWalk.set_gamma` and
            :meth:`hiperwalk.ContinuousWalk.set_marked`.

        Notes
        -----
        The Hamiltonian is given by

        .. math::
            H = -\gamma A  - \sum_{m \in M} \ket m \bra m,

        where :math:`A` is the adjacency matrix, and
        :math:`M` is the set of marked vertices.

        See Also
        --------
        set_gamma
        set_marked
        """

        # if laplacian:
        #     degrees = self.adj_matrix.sum(axis=1)
        #     H = scipy.sparse.diags(degrees, format="csr")
        #     del degrees
        #     H -= self.adj_matrix
        #     H *= -gamma

        gamma_kwargs = ContinuousWalk._filter_valid_kwargs(
                              kwargs,
                              ContinuousWalk._hamiltonian_kwargs['gamma'])
        marked_kwargs = ContinuousWalk._filter_valid_kwargs(
                              kwargs,
                              ContinuousWalk._hamiltonian_kwargs['marked'])

        self.set_gamma(**gamma_kwargs)
        self.set_marked(**marked_kwargs)

    def get_hamiltonian(self):
        r"""
        Returns the Hamiltonian.

        Returns
        -------
        :class:`scipy.sparse.csr_array`
        """
        if self._hamiltonian is not None:
            return self._hamiltonian

        H = -self._gamma * self._graph.adj_matrix

        # creating oracle
        if len(self._marked) > 0:
            data = np.ones(len(self._marked), dtype=np.int8)
            oracle = scipy.sparse.csr_array(
                    (data, (self._marked, self._marked)),
                    shape=(self.hilb_dim, self.hilb_dim))

            H -= oracle

        self._hamiltonian = H
        # since the hamiltonian was changed,
        # the previous evolution operator may not be coherent.
        self._evolution = None
        return H

    def set_evolution(self, **kwargs):
        r"""
        Alias for :meth:`hiperwalk.ContinuousWalk.set_hamiltonian`.
        """
        self.set_hamiltonian(**kwargs)

    def get_evolution(self, time=None, hpc=True):
        r"""
        Returns the evolution operator.

        Constructs the evolution operator based on the previously
        set Hamiltonian.

        Parameters
        ----------
        time : float
            Generates the evolution operator corresponding to the specified time.

        hpc : bool, default = True
            Determines whether or not to use neblina HPC 
            functions to generate the evolution operator.
        Returns
        -------
        :class:`numpy.ndarray`.

        Raises
        ------
        ValueError
            If ``time < 0``.

        See Also
        --------
        set_hamiltonian

        Notes
        -----
        The evolution operator is given by

        .. math::
            U = e^{-\text{i}tH},

        where :math:`H` is the Hamiltonian, and
        :math:`t` is the time.

        The evolution operator is constructed using
        a Taylor series expansion.

        .. warning::
            For non-integer time (floating number),
            the result is approximate. It is recommended 
            to select a small time interval and perform 
            multiple matrix multiplications to minimize 
            rounding errors.
        """
        if time is None or time < 0:
            raise ValueError(
                "Expected non-negative `time` value."
            )

        H = self.get_hamiltonian()

        if hpc and not self._pyneblina_imported():
            hpc = False

        if hpc:
            # determining the number of terms in power series
            max_val = np.max(np.abs(H))
            if max_val*time <= 1:
                nbl_U = nbl.matrix_power_series(
                        -1j*time*H, 30)

            else:
                # if the order of magnitude is very large,
                # float point errors may occur
                if ((isinstance(time, int) or time.is_integer())
                    and max_val <= 1
                ):
                    new_time = 1
                    num_mult = time - 1
                else:
                    new_time = max_val*time
                    order = np.ceil(np.math.log(new_time, 20))
                    new_time /= 10**order
                    num_mult = int(np.round(time/new_time)) - 1

                new_nbl_U = nbl.matrix_power_series(
                        -1j*new_time*H, 20)
                nbl_U = nbl.multiply_matrices(new_nbl_U, new_nbl_U)
                for i in range(num_mult - 1):
                    nbl_U = nbl.multiply_matrices(nbl_U, new_nbl_U)

            U = nbl.retrieve_matrix(nbl_U)

        else:
            U = scipy.linalg.expm(-1j*time*H.todense())

        self._evolution = U
        return U

    def simulate(self, time=None, initial_state=None,
                 initial_condition=None, hpc=True):
        r"""
        Analogous to :meth:`hiperwalk.QuantumWalk.simulate`,
        which accepts float entries for the ``time`` parameter.

        Parameters
        ----------
        time : float or tuple of floats
            This parameter is analogous to those in
            :meth:`hiperwalk.QuantumWalk.simulate`,
            with the distinction that it accepts float inputs.
            The ``step`` parameter is used to construct the evolution operator.
            The states within the interval
            **[** ``start/step``, ``end/step`` **]** are stored.
            The values describing this interval are
            rounded up if the decimal part exceeds ``1 - 1e-5``,
            and rounded down otherwise.

        Other Parameters
        ----------------
        See `hiperwalk.QuantumWalk.simulate`.

        See Also
        --------
        set_evolution
        get_evolution
        """
        if time is None:
            raise ValueError(
                "Invalid `time_range`. Expected a float, 2-tuple, "
                + "or 3-tuple of float."
            )

        time = np.array(self._time_to_tuple(time))

        self.get_evolution(time=time[2], hpc=hpc)

        # cleaning time_range to int
        tol = 1e-5
        time = [int(val/time[2])
                if int(val/time[2])
                   <= np.ceil(val/time[2]) - tol
                else int(np.ceil(val/time[2]))
                for val in time]

        states = super().simulate(time, initial_state,
                                  initial_condition, hpc)
        return states



# -*- coding: utf-8 -*-
"""
Created on Mon Jun 23 04:56:43 2014

@author: aaron
"""

import os
import subprocess
import numpy as np
import config as cfg

    
def runCore_DTQW1D():
    if cfg.TEST_MODE:
        os.system("neblina -id %d HIPERWALK_TEMP_DTQW1D.nbl %d %d %d %d %d %d %d >> /dev/null "%(cfg.HARDWAREID,cfg.STEPS,cfg.STATESIZE,cfg.GRAPHSIZE,cfg.DISTANCE_VECTOR_SIZE,cfg.SAVE_STATES_MULTIPLE_OF_N,cfg.ANIMATION,cfg.STATESIZE/2))
    else:
        os.system("neblina -id %d HIPERWALK_TEMP_DTQW1D.nbl %d %d %d %d %d %d %d"%(cfg.HARDWAREID,cfg.STEPS,cfg.STATESIZE,cfg.GRAPHSIZE,cfg.DISTANCE_VECTOR_SIZE,cfg.SAVE_STATES_MULTIPLE_OF_N,cfg.ANIMATION,cfg.STATESIZE/2))
        
def runCore_DTQW2D():
    if cfg.TEST_MODE:
        os.system("neblina -id %d HIPERWALK_TEMP_DTQW2D.nbl %d %d %d %d %d %d %d >> /dev/null "%(cfg.HARDWAREID,cfg.STEPS,cfg.STATESIZE,cfg.GRAPHSIZE,cfg.DISTANCE_VECTOR_SIZE,cfg.SAVE_STATES_MULTIPLE_OF_N,cfg.ANIMATION,cfg.STATESIZE/4))
    else:
        os.system("neblina -id %d HIPERWALK_TEMP_DTQW2D.nbl %d %d %d %d %d %d %d"%(cfg.HARDWAREID,cfg.STEPS,cfg.STATESIZE,cfg.GRAPHSIZE,cfg.DISTANCE_VECTOR_SIZE,cfg.SAVE_STATES_MULTIPLE_OF_N,cfg.ANIMATION,cfg.STATESIZE/4))

def runCore_STAGGERED1D():
    if cfg.TEST_MODE:
        os.system("neblina -id %d HIPERWALK_TEMP_STAGGERED1D.nbl %d %d %d %d %d %d >> /dev/null "%(cfg.HARDWAREID,cfg.STEPS,cfg.STATESIZE,cfg.GRAPHSIZE,cfg.DISTANCE_VECTOR_SIZE,cfg.SAVE_STATES_MULTIPLE_OF_N,cfg.ANIMATION))
    else:
        os.system("neblina -id %d HIPERWALK_TEMP_STAGGERED1D.nbl %d %d %d %d %d %d"%(cfg.HARDWAREID,cfg.STEPS,cfg.STATESIZE,cfg.GRAPHSIZE,cfg.DISTANCE_VECTOR_SIZE,cfg.SAVE_STATES_MULTIPLE_OF_N,cfg.ANIMATION))

def runCore_STAGGERED2D():
    if cfg.TEST_MODE:
        os.system("neblina -id %d HIPERWALK_TEMP_STAGGERED2D.nbl %d %d %d %d %d %d >> /dev/null "%(cfg.HARDWAREID,cfg.STEPS,cfg.STATESIZE,cfg.GRAPHSIZE,cfg.DISTANCE_VECTOR_SIZE,cfg.SAVE_STATES_MULTIPLE_OF_N,cfg.ANIMATION))
    else:
        os.system("neblina -id %d HIPERWALK_TEMP_STAGGERED2D.nbl %d %d %d %d %d %d"%(cfg.HARDWAREID,cfg.STEPS,cfg.STATESIZE,cfg.GRAPHSIZE,cfg.DISTANCE_VECTOR_SIZE,cfg.SAVE_STATES_MULTIPLE_OF_N,cfg.ANIMATION))


def runCore_CUSTOM():   
    a=" ".join([str(x) for x in cfg.CUSTON_OPERATORS_NAME])
    b=" ".join([str(x) for x in cfg.CUSTOM_LABELS_NAME])
    computeStates = 0
    
    if cfg.ADJMATRIX_PATH != "@NON_INICIALIZED@":
         computeStates = 1    
    

    if cfg.TEST_MODE:
        os.system("neblina -id %d HIPERWALK_TEMP_CUSTOM.nbl %d %d %d %d %d %d %d %d %d %s %s %s >> /dev/null"%(int(cfg.HARDWAREID),int(cfg.STEPS),int(cfg.SAVE_STATES_MULTIPLE_OF_N),int(cfg.SAVE_PROBS_MULTIPLE_OF_N), computeStates, int(cfg.ANIMATION),int(cfg.STATESIZE),int(cfg.GRAPHSIZE),int(cfg.CUSTOM_UNITARY_COUNTER),int(cfg.GRAPHDIMENSION),cfg.CUSTOM_INITIALSTATE_NAME,a,b))
    else:
        os.system("neblina -id %d HIPERWALK_TEMP_CUSTOM.nbl %d %d %d %d %d %d %d %d %d %s %s %s"%(int(cfg.HARDWAREID),int(cfg.STEPS),int(cfg.SAVE_STATES_MULTIPLE_OF_N),int(cfg.SAVE_PROBS_MULTIPLE_OF_N), computeStates, int(cfg.ANIMATION),int(cfg.STATESIZE),int(cfg.GRAPHSIZE),int(cfg.CUSTOM_UNITARY_COUNTER),int(cfg.GRAPHDIMENSION),cfg.CUSTOM_INITIALSTATE_NAME,a,b))
    
    
    
def generating_DTQW1D_NBL():
    output = open("HIPERWALK_TEMP_WALK.nbl",'w')    
    
    output.write("-- core.nbl\n")
    output.write("using io\n")
    output.write("using math\n")
    output.write("using time\n")
    
    output.write("def stdvr(OUTPUT_FILE, STEPS,PSI, DISTANCES, DISTANCES_SQUARE,ANIMATION,GRAPHSIZE)\n")
    output.write("    t0 = micro()\n")
    
    output.write("    p = vec_conj( PSI )\n")
#    output.write("    <<< p\n")
    output.write("    PROBABILITY_VECTOR  = vec_add_off( p,len(p)/2  )    \n")
#    output.write("   >>> PROBABILITY_VECTOR \n")
    output.write("    if ANIMATION == 1    \n")
    output.write("        f = open( \"NEBLINA_TEMP_PROB\" + tostr(STEPS, 5) + \".dat\", \"w\" )\n")
#    output.write("	 <<< PROBABILITY_VECTOR\n")
    output.write("        print( f, PROBABILITY_VECTOR )\n")
#    output.write("        >>> PROBABILITY_VECTOR\n")
    output.write("    end\n")
    output.write("\n")
    output.write("    tmp = vec_prod( DISTANCES_SQUARE, PROBABILITY_VECTOR )\n")
    output.write("\n")
    output.write("    s1  = vec_sum( tmp )    \n")
    output.write("    tmp = vec_prod( DISTANCES, PROBABILITY_VECTOR )\n")
    output.write("    mean = vec_sum( tmp )\n")
    output.write("    stddev = sqrt(abs(s1-mean*mean))\n")
    output.write("    println( OUTPUT_FILE, STEPS  + \"               \" + mean + \"               \" + s1 + \"               \" + stddev )\n")
    output.write("    t1 = micro()\n")
    output.write("--    println( t1 - t0 )\n")
    output.write("end\n")
    
    output.write("def main()\n")
#    
    output.write("fmtdouble( 16 )\n")
    output.write("   STATESIZE = %d\n"%cfg.STATESIZE)
#    
    output.write("   GRAPHSIZE= %d\n"%cfg.GRAPHSIZE)
#    
    output.write("   DISTANCE_VECTOR_SIZE = %d\n"%cfg.DISTANCE_VECTOR_SIZE)
#    
    output.write("   STEPS = %d\n"%cfg.STEPS)
#    
    output.write("   ANIMATION = %d\n"%cfg.ANIMATION)
#    
    output.write("   SAVE_STATES_MULTIPLE_OF_N=%d\n"%cfg.SAVE_STATES_MULTIPLE_OF_N)
#    
    output.write("   file_CtI = open( \"HIPERWALK_TEMP_COIN_TENSOR_IDENTITY_1D.dat\", \"r\" )\n")
    output.write("   CtI = sparse complex[STATESIZE,STATESIZE]\n")
    output.write("   read( file_CtI, CtI )\n")
#    
    output.write("   file_S = open( \"HIPERWALK_TEMP_SHIFT_OPERATOR_1D.dat\", \"r\" )\n")
    output.write("   S = sparse complex[STATESIZE,STATESIZE]\n")
    output.write("   read( file_S, S )\n")
#    
    output.write("   file_psi = open( \"HIPERWALK_TEMP_PSI.dat\", \"r\" )\n")
    output.write("   PSI = complex[STATESIZE]\n")
    output.write("   read( file_psi, PSI )\n")
#    
    output.write("   file_D = open(\"HIPERWALK_TEMP_DISTANCE_VECTOR.dat\", \"r\" ) \n")
    output.write("   DISTANCES_VECTOR = float[DISTANCE_VECTOR_SIZE ]\n")
    output.write("   read( file_D, DISTANCES_VECTOR )\n")
#    
#    output.write("   >>> DISTANCES_VECTOR  \n")
    output.write("   DISTANCES_SQUARE = vec_prod( DISTANCES_VECTOR,DISTANCES_VECTOR)\n")
#    output.write("   >>> CtI\n")
#    output.write("   >>> S\n")
#    output.write("   >>> PSI\n")
#
    output.write("   stdv_file = open(\"statistics.dat\", \"w\" ) \n")
    output.write("   println( stdv_file, \"#STEP               Mean                                      2nd Moment                                  Standard deviation\" )\n")
    
    output.write("   stdvr(stdv_file, 0, PSI,  DISTANCES_VECTOR  , DISTANCES_SQUARE, ANIMATION,GRAPHSIZE)\n")
#    
#
    output.write("   for t = 1 : STEPS\n")
    output.write("           PSI = mat_mulvec( CtI, PSI )\n")
    output.write("           PSI = mat_mulvec( S, PSI )\n")
#
#
    if cfg.ALLSTATES:
#        output.write("           <<< PSI\n")
        output.write("           if SAVE_STATES_MULTIPLE_OF_N > 0 \n")
        output.write("               a=t % SAVE_STATES_MULTIPLE_OF_N \n")
        output.write("               if a == 0 \n")
        output.write("                   f = open( \"wavefunction-\" + tostr(t) + \".dat\", \"w\" )\n")
        output.write("                   print( f, PSI )\n")
        output.write("                   stdvr(stdv_file, t, PSI,  DISTANCES_VECTOR  , DISTANCES_SQUARE, ANIMATION,GRAPHSIZE)\n")
        output.write("               end\n")
        output.write("           end\n")
#        output.write("           if SAVE_STATES_MULTIPLE_OF_N == 0 \n")
#        output.write("               f = open( \"wavefunction-\" + tostr(t) + \".dat\", \"w\" )\n")        
#        output.write("               print( f, PSI )\n")
#
#        output.write("           end\n")
#        output.write("           >>> PSI\n")
    else:
        output.write("           stdvr(stdv_file, t, PSI,  DISTANCES_VECTOR  , DISTANCES_SQUARE, ANIMATION,GRAPHSIZE)\n")

    output.write("   end  \n")
#    
    output.write("  PROBABILITY_VECTOR = vec_conj( PSI )\n")
    output.write("   <<< PSI, PROBABILITY_VECTOR \n")
    
    
    output.write("   final_prob = vec_add_off( PROBABILITY_VECTOR , len(PROBABILITY_VECTOR)/2 )\n")
    output.write("   println( \"[Neblina] Statistics file: .............. statistics.dat\" )\n")
    output.write("   final_state_neblina = open(\"NEBLINA_TEMP_final_state.dat\", \"w\")\n")
    output.write("   println( final_state_neblina, PSI )\n")
    output.write("   final_distribution_neblina = open(\"NEBLINA_TEMP_final_distribution.dat\", \"w\")\n")
    output.write("   println( final_distribution_neblina, final_prob )\n")
    
    output.write("   println( \"[Neblina] Wave Function final : .............. final_state.dat\" )\n")
    output.write("   println(\"[Neblina] Done!\")\n")
    output.write("end\n")
    output.close()

    

    




def generating_DTQW2D_NBL():

    output = open("HIPERWALK_TEMP_WALK.nbl",'w')    
    output.write(" -- core.nbl \n")
    output.write(" using io \n")
    output.write(" using std \n")
    output.write(" using math \n")
    output.write(" using time \n")

    output.write(" def stdv( OUTPUT_FILE, psi, x, x2, y, y2, ANIMATION, t ) \n")
    output.write("  \n")
    output.write("     prob = vec_conj( psi ) \n")
    output.write("     prob = vec_add_off( prob, len(prob)/4 ) \n")
    output.write("      \n")
    output.write("     if ANIMATION == 1 \n")
    output.write("         f = open( \"NEBLINA_TEMP_PROB\" + tostr(t, 5) + \".dat\", \"w\" ) \n")
    output.write(" 	    print( f, prob )     \n")
    output.write("     end \n")
    output.write("  \n")
    output.write("     vX = vec_prod( x, prob ) \n")
    output.write("     vY = vec_prod( y, prob ) \n")
    output.write("      \n")
    output.write("     vXX = vec_prod( x2, prob ) \n")
    output.write("     vYY = vec_prod( y2, prob ) \n")
    output.write("      \n")
    output.write("     sX = vec_sum( vX ) \n")
    output.write("     sY = vec_sum( vY ) \n")
    output.write("      \n")
    output.write("     sXX = vec_sum( vXX ) \n")
    output.write("     sYY = vec_sum( vYY ) \n")
    output.write("      \n")
    output.write("     varX = sXX - sX*sX \n")
    output.write("     varY = sYY - sY*sY \n")
    output.write("      \n")
    output.write("     println( OUTPUT_FILE, t  + \"            \" + sX + \"               \" + sY + \"                \" + varX + \"                \" + varY + \"              \" + sqrt( varX + varY ) ) \n")
#    output.write("     println( OUTPUT_FILE, t   \"         \"  sX  \"          \"  sY  \"          \"  varX  \"          \"  varY  \"          \"  sqrt( varX  varY ) ) \n")
    output.write("      \n")
    output.write(" end \n")
    output.write("  \n")

    output.write(" def main() \n")
    output.write("    fmtdouble( 16 ) \n")
    output.write("   STATESIZE = %d\n"%cfg.STATESIZE)
    output.write("   GRAPHSIZE= %d\n"%cfg.GRAPHSIZE)   
    output.write("   DISTANCE_VECTOR_SIZE = %d\n"%cfg.DISTANCE_VECTOR_SIZE)
    output.write("   STEPS = %d\n"%cfg.STEPS)
    output.write("   ANIMATION = %d\n"%cfg.ANIMATION)
    output.write("   SAVE_STATES_MULTIPLE_OF_N=%d\n"%cfg.SAVE_STATES_MULTIPLE_OF_N)
    output.write("    file_CtI = open( \"HIPERWALK_TEMP_COIN_TENSOR_IDENTITY_OPERATOR_2D.dat\", \"r\" ) \n")
    output.write("    CtI = sparse complex[STATESIZE,STATESIZE] \n")
    output.write("    read( file_CtI, CtI ) \n")
    output.write("    file_S = open( \"HIPERWALK_TEMP_SHIFT_OPERATOR_2D.dat\", \"r\" ) \n")
    output.write("    S = sparse complex[STATESIZE,STATESIZE] \n")
    output.write("    read( file_S, S ) \n")
    output.write("    file_psi = open( \"HIPERWALK_TEMP_PSI.dat\", \"r\" ) \n")
    output.write("    PSI = complex[STATESIZE] \n")
    output.write("    read( file_psi, PSI ) \n")
    output.write("     \n")
    output.write("    file_X = open(\"HIPERWALK_TEMP_X.dat\", \"r\" )  \n")
    output.write("    file_Y = open(\"HIPERWALK_TEMP_Y.dat\", \"r\" ) \n")
    output.write("    X = float[DISTANCE_VECTOR_SIZE] \n")
    output.write("    Y = float[DISTANCE_VECTOR_SIZE] \n")
    output.write("    read( file_X, X ) \n")
    output.write("    read( file_Y, Y ) \n")
    output.write("     \n")
    output.write("    X2 = vec_prod( X, X ) \n")
    output.write("    Y2 = vec_prod( Y, Y ) \n")
    output.write("     \n")
    output.write("    stdv_file = open(\"statistics.dat\", \"w\" )  \n")
    output.write("    println( stdv_file, \"#STEP     Mean(x)                                         Mean(y)                                                Var(x)                                                Var(y)                                                Standard deviation\" ) \n")
    output.write("    stdv( stdv_file, PSI, X, X2, Y, Y2, ANIMATION, 0 ) \n")
    output.write("    for t = 1 : STEPS \n")
    output.write("            PSI = mat_mulvec( CtI, PSI ) \n")
    output.write("            PSI = mat_mulvec( S, PSI ) \n")
    if cfg.ALLSTATES:
        output.write("            if SAVE_STATES_MULTIPLE_OF_N > 0  \n")
        output.write("                a=t % SAVE_STATES_MULTIPLE_OF_N  \n")
        output.write("                if a == 0  \n")
        output.write("                    f = open( \"wavefunction-\" + tostr(t) + \".dat\", \"w\" ) \n")
#        output.write("                    f = open( \"wavefunction-\"  tostr(t)  \".dat\", \"w\" ) \n")
        output.write("                    print( f, PSI ) \n")
        output.write("                stdv( stdv_file, PSI, X, X2, Y, Y2, ANIMATION, t ) \n")
        output.write("                end \n")

        output.write("            end \n")

        

    else:
        output.write("            stdv( stdv_file, PSI, X, X2, Y, Y2, ANIMATION, t ) \n")
    output.write("    end   \n")
    output.write("    PROBABILITY_VECTOR = vec_conj( PSI ) \n")
    output.write("    final_prob = vec_add_off( PROBABILITY_VECTOR , len(PROBABILITY_VECTOR)/4 ) \n")
    output.write("    println( \"[Neblina] Statistics file: .............. statistics.dat\" ) \n")
    output.write("    final_state_neblina = open(\"NEBLINA_TEMP_final_state.dat\", \"w\") \n")
    output.write("    println( final_state_neblina, PSI ) \n")
    output.write("    final_distribution_neblina = open(\"NEBLINA_TEMP_final_distribution.dat\", \"w\") \n")
    output.write("    println( final_distribution_neblina, final_prob ) \n")
    output.write("    println( \"[Neblina] Wave Function final : .............. final_state.dat\" ) \n")
    output.write("    println(\"[Neblina] Done!\") \n")
    output.write(" end \n")
    output.close()
    



   


    
def generating_STAGGERED1D_NBL():
    
    output = open("HIPERWALK_TEMP_WALK.nbl",'w')    
    
    ##
    ## Defining statistics
    ##    
    output.write("-- core.nbl\n")
    output.write("using io\n")
    output.write("using math\n")
    output.write("using time\n")
    output.write("def stdvr(OUTPUT_FILE, STEPS,PSI, DISTANCES, DISTANCES_SQUARE,ANIMATION,GRAPHSIZE)\n")
    output.write("    t0 = micro()\n")
    output.write("    PROBABILITY_VECTOR = vec_conj( PSI )\n")
    output.write("    if ANIMATION == 1    \n")
    output.write("        f = open( \"NEBLINA_TEMP_PROB\" + tostr(STEPS, 5) + \".dat\", \"w\" )\n")
    output.write("        print( f, PROBABILITY_VECTOR )\n")
    output.write("    end\n")
    
    output.write("    tmp = vec_prod( DISTANCES_SQUARE, PROBABILITY_VECTOR )\n")
    
    output.write("    s1  = vec_sum( tmp )    \n")
    output.write("    tmp = vec_prod( DISTANCES, PROBABILITY_VECTOR )\n")
    output.write("    mean = vec_sum( tmp )\n")
    output.write("    stddev = sqrt(abs(s1-mean*mean))\n")
    output.write("    println( OUTPUT_FILE, STEPS  + \"         \" + mean + \"          \" + s1 + \"          \" + stddev )\n")
    output.write("    t1 = micro()\n")
    output.write("--    println( t1 - t0 )\n")
    output.write("end\n")
    
    ##
    ## Defining main
    ##
    output.write("def main()\n")
    output.write("fmtdouble( 16 )\n")
    # Reading variables
    


    output.write("   STATESIZE = %d\n"%cfg.STATESIZE)
#    
    output.write("   STEPS = %d\n"%cfg.STEPS)
#    
    output.write("   ANIMATION = %d\n"%cfg.ANIMATION)

    output.write("   SAVE_STATES_MULTIPLE_OF_N=%d\n"%cfg.SAVE_STATES_MULTIPLE_OF_N)

    output.write("   file_Ue = open( \"HIPERWALK_TEMP_STAGGERED_EVEN_OPERATOR_1D.dat\", \"r\" )\n")
    output.write("   Ue = sparse complex[STATESIZE,STATESIZE]\n")
    output.write("   read( file_Ue, Ue )\n")
    output.write("   file_Uo = open( \"HIPERWALK_TEMP_STAGGERED_ODD_OPERATOR_1D.dat\", \"r\" )\n")
    output.write("   Uo = sparse complex[STATESIZE,STATESIZE]\n")
    output.write("   read( file_Uo, Uo )\n")
    output.write("   file_psi = open( \"HIPERWALK_TEMP_PSI.dat\", \"r\" )\n")
    output.write("   PSI = complex[STATESIZE]\n")
    output.write("   read( file_psi, PSI )\n")
    output.write("   file_D = open(\"HIPERWALK_TEMP_RANGE_1D.dat\", \"r\" ) \n")
    output.write("   DISTANCES_VECTOR = float[STATESIZE]\n")
    output.write("   read( file_D, DISTANCES_VECTOR )\n")

    # Sending variables to device
    output.write("   >>> DISTANCES_VECTOR  \n")
    output.write("   DISTANCES_SQUARE = vec_prod( DISTANCES_VECTOR,DISTANCES_VECTOR)\n")
    output.write("   stdv_file = open(\"statistics.dat\", \"w\" ) \n")
    output.write("   println( stdv_file, \"#STEP     Mean          2nd Moment    Standard deviation\" )\n")
    output.write("   stdvr(stdv_file, 0, PSI,  DISTANCES_VECTOR  , DISTANCES_SQUARE, ANIMATION,STATESIZE)\n")
    # Markov chain
    output.write("   for t = 1 : STEPS\n")
    output.write("           PSI = mat_mulvec( Ue, PSI )\n")
    output.write("           PSI = mat_mulvec( Uo, PSI )\n")

    if cfg.ALLSTATES:
        output.write("           if SAVE_STATES_MULTIPLE_OF_N > 0 \n")        
        output.write("               a=t % SAVE_STATES_MULTIPLE_OF_N \n")
        output.write("               if a == 0 \n")
        output.write("                   f = open( \"wavefunction-\" + tostr(t) + \".dat\", \"w\" )\n")
        output.write("                   print( f, PSI )\n")
        output.write("               end\n")
        output.write("           end\n")
        output.write("           if SAVE_STATES_MULTIPLE_OF_N == 0 \n")
        output.write("               f = open( \"wavefunction-\" + tostr(t) + \".dat\", \"w\" )\n")        
        output.write("               print( f, PSI )\n")
        output.write("           end\n")
    
    output.write("           stdvr(stdv_file, t, PSI,  DISTANCES_VECTOR  , DISTANCES_SQUARE, ANIMATION,STATESIZE)\n")
    output.write("   end  \n")
    
    output.write("   PROBABILITY_VECTOR = vec_conj( PSI )\n")
#    output.write("   final_prob = vec_add_off( PROBABILITY_VECTOR , len(PROBABILITY_VECTOR)/2 )\n")
#    output.write("   final_prob = vec_conj( PSI )\n")    
    output.write("   println( \"[Neblina] Statistics file: .............. statistics.dat\" )\n")
    output.write("   final_state_neblina = open(\"NEBLINA_final_state.dat\", \"w\")\n")
    output.write("   println( final_state_neblina, PSI )\n")
    output.write("   final_distribution_neblina = open(\"NEBLINA_final_distribution.dat\", \"w\")\n")
    output.write("   println( final_distribution_neblina, PROBABILITY_VECTOR )\n")
    output.write("   println( \"[Neblina] Wave Function final : .............. final_state.dat\" )\n")
    output.write("   println(\"[Neblina] Done!\")\n ")
    output.write("end")
    
    output.close()
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    

def neblina_state_to_vector(fileIn):
    qtdeValues = 0  
    for line in open(fileIn):  
        if line == "\n":  
            continue  
        if line.startswith("#"):  
            continue  
        qtdeValues = qtdeValues + 1  

    vector=np.zeros((qtdeValues,1),dtype=complex)
    actual = 0  
    
    for line in open(fileIn):  
        if line == "\n":  
            continue  
        if line.startswith("#"):  
            continue          
        line = line.split()  
        a=float(line[0])
        b=1J*float(line[1])
        vector[actual] = a+b
        actual = actual + 1  
    return vector


def neblina_distribution_to_vector(fileIn):
    qtdeValues = 0  
    for line in open(fileIn):  
        if line == "\n":  
            continue  
        if line.startswith("#"):  
            continue  
        qtdeValues = qtdeValues + 1  

    vector=np.zeros((qtdeValues,1),dtype=float)
    actual = 0  
    
    for line in open(fileIn):  
        if line == "\n":  
            continue  
        if line.startswith("#"):  
            continue          
        line = line.split()  
        a=float(line[0])
        vector[actual] = a
        actual = actual + 1  
    return vector






def generating_CUSTOM_NBL():
    output = open("walk.nbl",'w')    
    
    output.write("-- core.nbl\n")
    output.write("using io\n")
    output.write("using math\n")
    output.write("using time\n")
    
    output.write("def main()\n")
 
    output.write("   fmtdouble( 16 )\n")

    output.write("   STATESIZE = %d\n"%cfg.STATESIZE)
    
    output.write("   STEPS = %d\n"%cfg.STEPS)
    
    output.write("   SAVE_STATES_MULTIPLE_OF_N=%d\n"%cfg.SAVE_STATES_MULTIPLE_OF_N)

    counter=0
    for i in cfg.CUSTON_OPERATORS_NAME:
        output.write("   file_U%d = open( \"%s\", \"r\" )\n"%(counter,i))
        output.write("   U%d = sparse complex[STATESIZE,STATESIZE]\n"%counter)
        output.write("   read( file_U%d, U%d )\n"%(counter,counter))
        counter=counter+1
    
    output.write("   file_psi = open( \"%s\", \"r\" )\n"%cfg.CUSTOM_INITIALSTATE_NAME)
    output.write("   PSI = complex[STATESIZE]\n")
    output.write("   read( file_psi, PSI )\n")


#    for i in range(counter):
#        output.write("   >>> U%d\n"%i)    
#    output.write("   >>> PSI\n")   
    
    output.write("   for t = 1 : STEPS\n")

    for i in range(counter):
            output.write("           PSI = mat_mulvec( U%d, PSI )\n"%i)

    if cfg.ALLSTATES:
#        output.write("           <<< PSI\n")
        output.write("           if SAVE_STATES_MULTIPLE_OF_N > 0 \n")        
        output.write("               a=t % SAVE_STATES_MULTIPLE_OF_N \n")
        output.write("               if a == 0 \n")
        output.write("                   f = open( \"wavefunction-\" + tostr(t) + \".dat\", \"w\" )\n")
        output.write("                   print( f, PSI )\n")
        output.write("               end\n")
        output.write("           end\n")


#        output.write("           >>> PSI\n")
        
    output.write("   end  \n")
#    output.write("   <<< PSI\n")
    output.write("   final = open(\"final_state.dat\", \"w\")\n")
    output.write("   println( final, PSI)\n")
    output.write("   println(\"[Neblina] Done!\")\n")
    output.write("end\n")
    output.close()




import numpy as np
from scipy.sparse import csr_array
from sys import path as sys_path
from .graph import Graph

class Line(Graph):
    r"""
    Finite line graph (path graph).

    Parameters
    ----------
    num_vert : int
        The number of vertices on the line.

    Notes
    -----
    In the :obj:`Line` class, directions can be assigned to the arcs. 
    An arc pointing to the right has direction 0 (e.g., (1, 2)), 
    and an arc pointing to the left has direction 1 (e.g., (2, 1)).

    The order of the arcs is determined by their direction. 
    Thus, for a vertex :math:`v \in V`, 
    the arcs :math:`(v, v + 1)` and :math:`(v, v - 1)` 
    have labels :math:`a_0` and :math:`a_1` respectively, 
    with :math:`a_0 < a_1`. 
    The only exceptions to this rule are the extreme vertices 0 
    and :math:`|V| - 1`, as they have outdegree 1. 

    Apart from these exceptions, 
    for any two vertices :math:`v_1 < v_2`, 
    any arc with tail :math:`v_1` will have a label smaller than the 
    label of any arc with tail :math:`v_2`.

    For instance, Figure 1 illustrates the labels of the arcs of a path graph with 4 vertices.
    
    .. graphviz:: ../../graphviz/line-arcs.dot
        :align: center
        :layout: neato
        :caption: Figure 1.
    """

    def __init__(self, num_vert):
        # Creating adjacency matrix
        # The end vertices are only adjacent to 1 vertex.
        # Every other vertex is adjacent to two vertices.
        data = np.ones(2*(num_vert-1), dtype=np.int8)
        # upper diagonal
        row_ind = [lin+shift for shift in range(2)
                             for lin in range(num_vert - 1)]
        # lower digonal
        col_ind = [col+shift for shift in [1, 0]
                             for col in range(num_vert - 1)]
        adj_matrix = csr_array((data, (row_ind, col_ind)))
    
        # initializing
        super().__init__(adj_matrix)

    def embeddable(self):
        return True

    def default_coin(self):
        r"""
        Returns the default coin name.

        The default coin for a coined quantum walk on a
        segment is ``'hadamard'``.
        """
        return 'hadamard'

    def arc_label(self, tail, head):
        diff = head - tail
        if diff != 1 and diff != -1:
            raise ValueError('Invalid arc.')

        num_vert = self.number_of_vertices()
        return (2*tail - 1 if (diff == 1 and tail != 0
                               or tail == num_vert - 1)
                else tail*2)

    def arc(self, label):
        label += 1
        tail = label//2
        num_vert = self.number_of_vertices()
        head = (tail + (-1)**(label % 2)
                if tail != 0 and tail != num_vert - 1
                else tail - (-1)**(label % 2))
        return (tail, head)

    def next_arc(self, arc):
        # implemented only if is embeddable
        try:
            tail, head = arc
            diff = head - tail
            if diff != 1 and diff != -1:
                raise ValueError('Invalid arc')

            if head == 0 or head == self.number_of_vertices() - 1:
                return (head, tail)
            
            return ((tail + 1, head + 1) if diff == 1
                    else (tail - 1, head - 1))
        except TypeError:
            if arc == 0:
                return 1
            num_arcs = self.number_of_arcs()
            if arc == num_arcs - 1:
                return num_arcs - 2
            return arc + 2 if arc % 2 == 1 else arc - 2

    def previous_arc(self, arc):
        # implemented only if is embeddable
        try:
            tail, head = arc
            diff = head - tail
            if diff != 1 and diff != -1:
                raise ValueError('Invalid arc')

            if tail == 0 or tail == self.number_of_vertices() - 1:
                return (head, tail)
            
            return ((tail - 1, head - 1) if diff == 1
                    else (tail + 1, head + 1))
        except TypeError:
            num_arcs = self.number_of_arcs()
            if arc < 0 or arc >= num_arcs:
                raise ValueError('Invalid arc')

            arc = arc + 2 if arc % 2 == 0 else arc - 2
            if arc < 0:
                arc += 1
            elif arc >= num_arcs:
                arc -= 1

            return arc



import numpy as np
from sys import path as sys_path
sys_path.append('../')
sys_path.append('../../')
import hiperwalk as hpw
import unittest
from test_constants import *

class TestCoinedLine(unittest.TestCase):

    def setUp(self):
        self.num_vert = 41
        self.line = hpw.Line(self.num_vert)
        self.qw = hpw.CoinedWalk(self.line)

    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_persistent_shift_right_state_transfer(self):
        # initial state in leftmost vertex
        # final state in rightmost vertex
        self.qw.set_shift('persistent')
        self.qw.set_coin('I')
        self.qw.set_marked([])

        init_state = self.qw.state([1, (0, 1)])

        num_steps = self.num_vert - 1
        final_state = self.qw.simulate(num_steps, init_state, hpc=False)
        final_state = final_state[0]

        self.assertTrue(
            final_state[-1] == 1 and np.all(final_state[:-1] == 0)
        )

    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_persistent_shift_left_state_transfer(self):
        # initial state in leftmost vertex
        # final state in rightmost vertex
        self.qw.set_shift('persistent')
        self.qw.set_coin('I')
        self.qw.set_marked([])

        init_state = self.qw.state(
            [1, (self.num_vert - 1, self.num_vert - 2)])

        num_steps = self.num_vert - 1
        final_state = self.qw.simulate(num_steps, init_state, hpc=False)
        final_state = final_state[0]

        self.assertTrue(
            final_state[0] == 1 and np.all(final_state[1:] == 0)
        )

    @unittest.skipIf(not TEST_HPC, 'Skipping hpc tests.')
    def test_hpc_default_evolution_operator(self):

        num_steps = self.num_vert // 2
        center = self.num_vert // 2
        entries = [[1, (center, center + 1)],
                   [-1j, (center, center - 1)]]
        init_state = self.qw.state(*entries)

        states = self.qw.simulate((num_steps, 1), init_state, hpc=False)
        hpc_states = self.qw.simulate((num_steps, 1), init_state, hpc=True)

        self.assertTrue(
            np.allclose(states, hpc_states, rtol=1e-15, atol=1e-15)
        )



import numpy as np
from sys import path as sys_path
sys_path.append('../')
sys_path.append('../../')
from test_constants import *
import hiperwalk as hpw
import unittest

class TestCoinedCycle(unittest.TestCase):

    def setUp(self):
        self.num_vert = 10
        self.cycle = hpw.Cycle(self.num_vert)
        self.qw = hpw.CoinedWalk(self.cycle)


    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_clockwise_roundabout(self):
        self.qw.set_shift('persistent')
        self.qw.set_coin('I')
        self.qw.set_marked([])

        init_state = self.qw.state([1, (0, 1)])

        num_steps = self.num_vert
        final_state = self.qw.simulate(num_steps, init_state, hpc=False)[0]

        self.assertTrue(np.all(init_state == final_state))

    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_anticlockwise_roundabout(self):
        self.qw.set_shift('persistent')
        self.qw.set_coin('I')
        self.qw.set_marked([])

        init_state = self.qw.state([1, (0, self.num_vert - 1)])

        num_steps = self.num_vert
        final_state = self.qw.simulate(num_steps, init_state, hpc=False)[0]

        self.assertTrue(np.all(init_state == final_state))

    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_hadamard_evolution_operator(self):
        init_state = self.qw.state((1, (0, 1)))

        num_steps = 2*self.num_vert
        states = self.qw.simulate((num_steps, 1), init_state, hpc=False)
        states = states.real

        def recursive_expression():
            states = np.zeros((num_steps + 1, init_state.shape[0]))
            states[0] = np.copy(init_state)
            num_arcs = 2*self.num_vert

            for i in range(1, num_steps + 1):
                for j in range(init_state.shape[0]):
                    if j % 2 == 0:
                        states[i][j] = (states[i-1][(j - 2) % num_arcs]
                                        + states[i-1][(j - 1) % num_arcs]
                                       ) / np.sqrt(2)
                    else:
                        states[i][j] = (states[i-1][(j + 1) % num_arcs]
                                        - states[i-1][(j + 2) % num_arcs]
                                       ) / np.sqrt(2)
            return states


        rec_states = recursive_expression()
        
        self.assertTrue(np.allclose(states, rec_states,
                                    rtol=1e-15, atol=1e-15))

    @unittest.skipIf(not TEST_HPC, 'Skipping hpc tests.')
    def test_hpc_clockwise_roundabout(self):
        self.qw.set_shift('persistent')
        self.qw.set_coin('I')
        self.qw.set_marked([])

        init_state = self.qw.state([1, (0, 1)])

        num_steps = self.num_vert
        final_state = self.qw.simulate(num_steps, init_state, hpc=True)[0]

        self.assertTrue(np.all(init_state == final_state))

    @unittest.skipIf(not TEST_HPC, 'Skipping hpc tests.')
    def test_hpc_anticlockwise_roundabout(self):
        self.qw.set_shift('persistent')
        self.qw.set_coin('I')
        self.qw.set_marked([])

        init_state = self.qw.state([1, (0, self.num_vert - 1)])

        num_steps = self.num_vert
        final_state = self.qw.simulate(num_steps, init_state, hpc=True)[0]

        self.assertTrue(np.all(init_state == final_state))

    @unittest.skipIf(not TEST_HPC, 'Skipping hpc tests.')
    def test_hpc_evolution_operator_matches_nonhpc(self):
        init_state = self.qw.state([1, (0, 1)])
        num_steps = 2*self.num_vert

        states = self.qw.simulate(
            (num_steps, 1), init_state, hpc=False)
        hpc_states = self.qw.simulate(
            (num_steps, 1), init_state, hpc=True)
        
        self.assertTrue(np.allclose(states, hpc_states,
                                    rtol=1e-15, atol=1e-15))



import numpy as np
import scipy.sparse
from .graph import Graph
from warnings import warn

class Lattice(Graph):
    r"""
    Two-dimensionsal lattice.

    The lattice can be designed with either
    cyclic boundary conditions or borders.
    Moreover, the lattice's representation can be either
    *natural* or *diagonal*.
    In the *natural* representation,
    neighboring vertices lie along the X and Y axes,
    while in the *diagonal* representation,
    they lie along the diagonals.

    Parameters
    ----------
    dimensions : int or tuple of int
        Lattice dimensions in ``(x_dim, y_dim)`` format.
        If ``dimensions`` is an integer, creates a square lattice.

    periodic : bool, default=True
        ``True`` if the lattice has cyclic boundary conditions,
        ``False`` if it has borders.

    diagonal : bool, default=False
        ``True`` if the lattice has the diagonal representation,
        ``False`` if it has the natural representation.

    Notes
    -----
    The order of the arcs is determined according to
    the order of the vertices.
    The order of the vertices is defined as follows:
    if :math:`(x_1, y_1)` and :math:`(x_2, y_2)` are two valid vertices,
    we say that :math:`(x_1, y_1) < (x_2, y_2)` if :math:`y_1 < y_2` or
    if :math:`y_1 = y_2` and :math:`x_1 < x_2`.
    The order of the arcs also depends on the lattice representation
    (natural or diagonal).

    Natural Lattice:
        In the natural lattice,
        the directions are described as follows:

        * 00 = 0: right;
        * 01 = 1: left;
        * 10 = 2: up;
        * 11 = 3: down.
        
        The most significant bit corresponds to the axis:
        0 represents the X-axis and 1 represents the Y-axis.
        The least significant bit indicates the direction of
        movement along the given axis, with 0 signifying
        forward movement and 1 signifying backward movement.

        The order of arcs corresponds with the order of vertices
        and their respective directions. For example,
        consider a vertex :math:`(x, y)`. Then,
        :math:`(x \pm 1, y)` and :math:`(x, y \pm 1)`
        are adjacent vertices.
        The order of these arcs is

        .. math::
            ((x, y), (x + 1, y)) &< ((x, y), (x - 1, y)) \\
                                 &< ((x, y), (x, y + 1)) \\
                                 &< ((x, y), (x, y - 1)).

        The directions are depicted in :ref:`fig-natural-dir`.

        .. graphviz:: ../../graphviz/lattice/natural-directions.dot
            :align: center
            :layout: neato
            :name: fig-natural-dir
            :caption: Figure: Directions in the natural representation.

        For example, the labels of the arcs for
        the :math:`3 \times 3`-lattice with periodic boundary
        conditions are depicted in :ref:`fig-periodic-natural-lattice`.

        .. graphviz:: ../../graphviz/lattice/periodic-natural.dot
            :align: center
            :layout: neato
            :name: fig-periodic-natural-lattice
            :caption: Figure: 3x3-lattice with cyclic boundary conditions.

        In the case of a natural lattice with borders, the labels of the
        arcs maintain the same sequence but with some modifications
        due to the presence of vertices with degrees 2 and 3.
        Figure :ref:fig-bounded-natural-Lattice provides an illustration
        of a bounded natural lattice.

        .. graphviz:: ../../graphviz/lattice/bounded-natural.dot
            :align: center
            :layout: neato
            :name: fig-bounded-natural-lattice
            :caption: Figure: Bounded natural 3x3-lattice.

    Diagonal Lattice:
        In the diagonal lattice,
        the directions are described as follows:

        * 00 = 0: right, up;
        * 01 = 1: right, down;
        * 10 = 2: left, up;
        * 11 = 3: left, down.

        Each binary value indicates the direction of movement,
        with 0 representing forward motion and 1 representing
        backward motion. The most significant bit corresponds to
        movement along the X-axis, while the least significant bit
        corresponds to movement along the Y-axis.

        The order of arcs corresponds with the order of vertices
        and their respective directions. For example,
        consider a vertex :math:`(x, y)`. Then,
        :math:`(x \pm 1, y)` and :math:`(x, y \pm 1)`
        are adjacent vertices.
        The order of these arcs is
        
        .. math::
            ((x, y), (x + 1, y + 1)) &< ((x, y), (x + 1, y - 1)) \\
                                     &< ((x, y), (x - 1, y + 1)) \\
                                     &< ((x, y), (x - 1, y - 1)).

        The directions are depicted in :ref:`fig-diagonal-dir`.

        .. graphviz:: ../../graphviz/lattice/diagonal-directions.dot
            :align: center
            :layout: neato
            :name: fig-diagonal-dir
            :caption: Figure: Directions in the diagonal representation.

        For example, the labels of the arcs for
        the :math:`3 \times 3`-lattice with periodic boundary
        conditions are depicted in :ref:`fig-periodic-diagonal-lattice`

        .. graphviz:: ../../graphviz/lattice/periodic-diagonal.dot
            :align: center
            :layout: neato
            :name: fig-periodic-diagonal-lattice
            :caption: Figure: Periodic diagonal 3x3-lattice.

        In the case of a diagonal lattice with borders, the labels of the
        arcs maintain the same sequence but with some modifications.
        Figure :ref:`fig-bounded-diagonal-Lattice` provides
        an illustration of a bounded diagonal lattice.

        .. graphviz:: ../../graphviz/lattice/bounded-diagonal.dot
            :align: center
            :layout: neato
            :name: fig-bounded-diagonal-lattice
            :caption: Figure: Bounded 3x3-lattice in the diagonal representation.

        Note that in this context,
        there exist two independent sublattices.
        In other words, a vertex in one sublattice is not accessible from
        a vertex in the other sublattice. This situation also arises
        if the diagonal lattice has periodic boundary conditions and both
        dimensions are even.
        Figure :ref:`fig-even-dim-diagonal` illustrates
        an example of this case.

        .. graphviz:: ../../graphviz/lattice/even-dim-diagonal.dot
            :align: center
            :layout: neato
            :name: fig-even-dim-diagonal
            :caption: Figure: 4x4-lattice with cyclic boundary conditions.
    """
    def __init__(self, dimensions, periodic=True, diagonal=False):
        try:
            x_dim, y_dim = dimensions
        except TypeError:
            x_dim = y_dim = dimensions

        num_vert = x_dim * y_dim
        num_arcs = (4*num_vert if periodic
                    else 4*num_vert - 2*(x_dim + y_dim))
        num_edges = num_arcs >> 1

        data = np.ones(num_arcs, np.int8)
        indptr = np.zeros(num_vert + 1)
        indices = np.zeros(num_arcs)
        arc_count = 0

        if not periodic and not diagonal:
            for v in range(num_vert):
                #indptr
                indptr[v + 1] = indptr[v] + 4
                x = v % x_dim
                y = v // x_dim
                if x == 0 or x == x_dim - 1:
                    indptr[v + 1] -= 1
                if y == 0 or y == y_dim - 1:
                    indptr[v + 1] -= 1

                #indices
                for coin in [3, 1, 0, 2]: #stay in order
                    head = v + (-1)**(coin % 2) * x_dim**(coin // 2)

                    if (head >= 0 and head < num_vert
                        and not (head % x_dim == 0 and head - v == 1)
                        and not (v % x_dim == 0 and v - head == 1)
                    ):
                        indices[arc_count] = head
                        arc_count += 1

        if not periodic and diagonal:
            raise NotImplementedError

        if periodic:
            indptr = np.arange(0, num_arcs + 1, 4)

            for v in range(num_vert):
                cols = (np.array(
                            [v - v % x_dim
                             + (v % x_dim + (-1)**(coin % 2)) % x_dim
                             if coin < 2 else
                             (v + x_dim*(-1)**(coin % 2)) % num_vert
                             for coin in range(4)])
                        if not diagonal
                        else np.array(
                            [((v % x_dim + (-1)**(coin // 2)) % x_dim
                              + x_dim*(v // x_dim + (-1)**(coin % 2)))
                             % num_vert
                             for coin in range(4)]))
                cols.sort()
                indices[arc_count:arc_count + 4] = cols
                arc_count += 4

            # TODO: use spdiags

        adj_matrix = scipy.sparse.csr_array((data, indices, indptr),
                                            shape=(num_vert, num_vert))
        super().__init__(adj_matrix)
        self.x_dim = x_dim
        self.y_dim = y_dim
        self.periodic = periodic
        self.diagonal = diagonal

    def embeddable(self):
        return True

    def vertex_coordinates(self, label):
        r"""
        Returns vertex (x, y)-coordinates given its label.

        Parameters
        ----------
        label : int
            Vertex label.

        Returns
        -------
        x : int
            Vertex X-coordinate.
        y : int
            Vertex Y-coordinate.

        See Also
        --------
        vertex_label
        """
        return (label % self.x_dim, label // self.x_dim)


    def vertex_label(self, x, y):
        r"""
        Returns vertex label (number) given its coordinates.

        Parameters
        ----------
        x : int
            Vertex X-coordinate.
        y : int
            Vertex Y-coordinate.

        Returns
        -------
        int
            Vertex label.

        See Also
        --------
        vertex_coordinates
        """
        return (x + self.x_dim*y) % self.number_of_vertices()

    def arc_direction(self, arc):
        r"""
        Return arc direction.

        Parameters
        ----------
        arc
            Any of the following notations are acceptable.
            
            * ((int, int), (int, int))
                Arc notation with vertices' coordinates.
            * (int, int)
                Arc notation with vertices' labels.
            * int
                Arc label.

        Returns
        -------
        int
            If natural (not diagonal) lattice:
                * 0: right
                * 1: left
                * 2: up
                * 3: down

            If diagonal lattice:
                * 0: right, up
                * 1: right, down
                * 2: left, up
                * 3: left, down

        Notes
        -----
        Does not check if arc exists.
        """
        # dealing with coordinates
        try:
            tail, head = arc
            if not hasattr(tail, '__iter__'):
                tail = self.vertex_coordinates(tail)
            if not hasattr(head, '__iter__'):
                head = self.vertex_coordinates(head)
        except TypeError:
            tail, head = self.arc(arc)

        if self.diagonal:
            x_diff = head[0] - tail[0]
            y_diff = head[1] - tail[1]
            x = 0 if (x_diff == 1 or x_diff == -self.x_dim + 1) else 1
            y = 0 if (y_diff == 1 or y_diff == -self.y_dim + 1) else 1
            return (x << 1) + y

        y = tail[1] != head[1]
        x = ((tail[1] - head[1]) % self.y_dim == 1
             if y else
             (tail[0] - head[0]) % self.x_dim == 1)

        return (y << 1) + x


    def arc_label(self, tail, head):
        try:
            tail = self.vertex_label(tail[0], tail[1])
            head = self.vertex_label(head[0], head[1])
        except TypeError:
            pass
        except IndexError:
            pass

        if self.adj_matrix[tail, head] == 0:
            raise ValueError('Inexistent arc ' + str((tail, head)) + '.')

        if self.periodic:
            return 4*tail + self.arc_direction((tail, head))

        label = self.adj_matrix.indptr[tail]
        direction = self.arc_direction((tail, head))
        if self.diagonal:
            raise NotImplementedError

        sub_x = (1 if ((tail % self.x_dim == 0
                        or tail % self.x_dim == self.x_dim - 1)
                       and direction > 0)
                 else 0)
            
        sub_y = (1 if ((tail // self.x_dim == 0
                        or tail // self.x_dim == self.x_dim - 1)
                       and direction > 2)
                 else 0)
        return label + direction - sub_x - sub_y

    def arc(self, label, coordinates=True):
        r"""
        Arc in arc notation.

        Given the arc label (a number),
        returns the arc in the ``(tail, head)`` notation.

        Parameters
        ----------
        label : int
            Arc label (number)
        coordinates : bool, default=True
            Whether the vertices are returned as coordinates or as labels.

        Returns
        -------
        (tail, head)
            There are two possible formats for the vertices
            ``tail`` and ``head``.

            (vertex_x, vertex_y) : (int, int)
                If ``coordinates=True``.
            label : int
                If ``coordinates=False``.
        """
        if not self.periodic and self.diagonal:
            raise NotImplementedError

        if self.periodic:
            tail = label // 4
            coin = label % 4
            num_vert = self.number_of_vertices()
            x_dim = self.x_dim
            if self.diagonal:
                head = (((tail % x_dim + (-1)**(coin // 2)) % x_dim
                          + x_dim*(tail // x_dim + (-1)**(coin % 2)))
                        % num_vert)
            else:
                head = (tail - tail % x_dim
                        + (tail % x_dim + (-1)**(coin % 2)) % x_dim
                        if coin < 2 else
                        (tail + x_dim*(-1)**(coin % 2)) % num_vert)

        else:
            # not diagonal
            tail, _ = super().arc(label)
            diff = label - self.adj_matrix.indptr[tail]
            num_vert = self.number_of_vertices()

            for coin in range(4):
                head = tail + (-1)**(coin % 2) * self.x_dim**(coin // 2)

                if (head >= 0 and head < num_vert
                    and not (head % self.x_dim == 0 and head - tail == 1)
                    and not (tail % self.x_dim == 0 and tail - head == 1)
                ):
                    if diff == 0:
                        break
                    diff -= 1

        arc = (tail, head)

        if not coordinates:
            return arc
        return (self.vertex_coordinates(arc[0]),
                self.vertex_coordinates(arc[1]))

    def neighbors(self, vertex):
        iterable = hasattr(vertex, '__iter__')
        if iterable:
            vertex = self.vertex_label(vertex[0], vertex[1])

        neigh = super().neighbors(vertex)

        if iterable:
            return list(map(self.vertex_coordinates, neigh))
        return neigh

    def arcs_with_tail(self, tail):
        try:
            tail = self.vertex_label(tail[0], tail[1])
        except TypeError:
            pass

        return super().arcs_with_tail(tail)

    def degree(self, vertex):
        try:
            vertex = self.vertex_label(vertex[0], vertex[1])
        except TypeError:
            pass

        return super().degree(vertex)

    def next_arc(self, arc):
        try:
            tail, head = arc
        except TypeError:
            tail, head = self.arc(arc)

        iterable = hasattr(tail, '__iter__')
        
        if not iterable:
            tail = self.vertex_coordinates(tail)
            head = self.vertex_coordinates(head)

        # get direction
        direction = self.arc_direction(arc)

        if self.diagonal:
            if not self.periodic:
                raise NotImplementedError

            tail = head
            x = direction // 2
            y = direction % 2
            head = ((head[0] + (-1)**x) % self.x_dim,
                    (head[1] + (-1)**y) % self.y_dim)
        else:
            y_axis = direction // 2
            exp = direction % 2
            if self.periodic:
                tail = head
                head = ((head[0], (head[1] + (-1)**exp) % self.x_dim)
                        if y_axis else
                        ((head[0] + (-1)**exp) % self.y_dim, head[1]))
            else:
                new_head = ((head[0], head[1] + (-1)**exp) if y_axis
                            else (head[0] + (-1)**exp, head[1]))

                if (new_head[0] < 0 or new_head[0] >= self.x_dim
                    or new_head[1] < 0 or new_head[1] >= self.y_dim
                ):
                    # out of bounds. Rebound
                    new_head = tail

                tail = head
                head = new_head

        if not iterable:
            tail = self.vertex_label(tail[0], tail[1])
            head = self.vertex_label(head[0], head[1])
        return (tail, head)

    def previous_arc(self, arc):
        arc_iterable = hasattr(arc, '__iter__')
        if arc_iterable:
            tail, head = arc
        else:
            tail, head = self.arc(arc)

        vertex_iterable = hasattr(tail, '__iter__')
        if not vertex_iterable:
            tail = self.vertex_coordinates(tail)
            head = self.vertex_coordinates(head)

        direction = self.arc_direction(arc)
        if self.diagonal:
            if not self.periodic:
                raise NotImplementedError

            x = direction // 2
            y = direction % 2
            head = tail
            tail = ((tail[0] - (-1)**x) % self.x_dim,
                    (tail[1] - (-1)**y) % self.y_dim)
        else:
            y_axis = direction // 2
            exp = direction % 2
            if self.periodic:
                head = tail
                tail = ((tail[0], (tail[1] - (-1)**exp) % self.x_dim)
                        if y_axis else
                        ((tail[0] - (-1)**exp) % self.y_dim, tail[1]))
            else:
                new_tail = ((tail[0], tail[1] - (-1)**exp) if y_axis
                            else (tail[0] - (-1)**exp, tail[1]))

                if (new_tail[0] < 0 or new_tail[0] >= self.x_dim
                    or new_tail[1] < 0 or new_tail[1] >= self.y_dim
                ):
                    # out of bounds. Rebound
                    new_tail = head

                head = tail
                tail = new_tail

        if not vertex_iterable:
            tail = self.vertex_label(tail[0], tail[1]),
            head = self.vertex_label(head[0], head[1])

        if not arc_iterable:
            return self.arc_label(tail, head)
        return (tail, head)

    def dimensions(self):
        r"""
        Lattice dimensions.

        Returns
        -------
        x_dim : int
            Dimension alongside de X axis.
        y_dim : int
            Dimension alongside de Y axis.
        """
        return (self.x_dim, self.y_dim)

    def get_central_vertex(self):
        r"""
        Vertex with label in the center of the graph as plane.

        .. deprecated:: 2.0a1
            ``get_central_vertex`` will be removed in Hiperwalk 2.0 because
            the user can calculate the central vertex easily using
            :meth:`dimensions`.

        The central vertex is the vertex that would be located at the
        plane center after mapping every vertex ``(x, y)`` to its
        respetive plane point.
        
        This is not the center vertex.

        Raises
        ------
        ValueError
            If any lattice dimension is even.
        """
        warn('`get_central_vertex` is deprecated. '
             + 'It will be removed in version 2.0.')

        if self.x_dim % 2 != 1 or self.y_dim % 2 != 1:
            raise ValueError(
                "One of lattice dimensions is even. "
                + "Hence it does not have a single central vertex."
            )
        return np.array([self.x_dim // 2, self.y_dim // 2])



import numpy as np
import scipy
import scipy.sparse
import networkx as nx
from .quantum_walk import QuantumWalk
from .._constants import __DEBUG__, PYNEBLINA_IMPORT_ERROR_MSG
from scipy.linalg import hadamard, dft
try:
    from . import _pyneblina_interface as nbl
except:
    pass

if __DEBUG__:
    from time import time as now

class CoinedWalk(QuantumWalk):
    r"""
    Manage an instance of a coined quantum walk on
    any simple graph.

    This class provides methods for handling, simulating, and generating
    operators within the coined quantum walk model for
    various types of graphs.
    
    For additional details about coined quantum walks,
    refer to the Notes Section.

    Parameters
    ----------
    graph
        Graph on which the quantum walk takes place.
        It can be the graph itself (:class:`hiperwalk.graph.Graph`) or
        its adjacency matrix (:class:`scipy.sparse.csr_array`).

    adjacency : :class:`scipy.sparse.csr_array`, optional
        .. deprecated:: 2.0a1
            It will be removed in version 2.0.
            Use ``graph`` instead.

        Use ``graph`` instead.
        Adjacency matrix of the graph.

    **kwargs : optional
        Optional arguments for setting the non-default evolution operator.
        See :meth:`set_evolution`.

    Warns
    -----
    If ``adjacency`` is set. It is deprecated. Use ``graph`` instead.

    Raises
    ------
    TypeError
        if ``adj_matrix`` is not an instance of
        :class:`scipy.sparse.csr_array`.

    See Also
    --------
    set_evolution

    Notes
    -----

    The computational basis is spanned by the set of arcs.
    The cardinality is :math:`2|E|`, where :math:`E`
    represents the edge set of the graph.
    For further information regarding the order of the arcs,
    consult the respective graph descriptions.
    The Hilbert space of a coined quantum walk is denoted by
    :math:`\mathcal{H}^{2|E|}`.

    For a more detailed understanding of coined quantum walks,
    refer to Section 7.2: Coined Walks on Arbitrary Graphs,
    found in the book  'Quantum Walks and Search Algorithms' [1]_.

    References
    ----------
    .. [1] R. Portugal. "Quantum walks and search algorithms", 2nd edition,
        Springer, 2018.
    """

    # static attributes.
    # The class must be instantiated once, otherwise are dicts are empty.
    # The class must be instantiated so the interpreter knows
    # the memory location for the function pointers.
    _coin_funcs = dict()
    _valid_kwargs = dict()

    def __init__(self, graph=None, adjacency=None, **kwargs):

        self._shift = None
        self._coin = None
        self._oracle_coin = []
        super().__init__(graph=graph, adjacency=adjacency)

        # Expects adjacency matrix with only 0 and 1 as entries
        self.hilb_dim = self._graph.number_of_arcs()

        if not bool(CoinedWalk._valid_kwargs):
            # assign static attribute
            CoinedWalk._valid_kwargs = {
                'shift': CoinedWalk._get_valid_kwargs(self.set_shift),
                'coin': CoinedWalk._get_valid_kwargs(self.set_coin),
                'marked': CoinedWalk._get_valid_kwargs(self.set_marked)
            }

        # dict with valid coins as keys and the respective
        # function pointers.
        if not bool(CoinedWalk._coin_funcs):
            # assign static attribute
            CoinedWalk._coin_funcs = {
                'fourier': CoinedWalk._fourier_coin,
                'grover': CoinedWalk._grover_coin,
                'hadamard': CoinedWalk._hadamard_coin,
                'identity': CoinedWalk._identity_coin,
                'minus_fourier': CoinedWalk._minus_fourier_coin,
                'minus_grover': CoinedWalk._minus_grover_coin,
                'minus_hadamard': CoinedWalk._minus_hadamard_coin,
                'minus_identity': CoinedWalk._minus_identity_coin
            }

        self.set_evolution(**kwargs)

        if __DEBUG__:
            methods = list(self._valid_kwargs)
            params = [p for m in methods
                        for p in self._valid_kwargs[m]]
            if len(params) != len(set(params)):
                raise AssertionError


    def _set_flipflop_shift(self):
        r"""
        Create the flipflop shift operator (:math:`S`) based on
        the ``_graph`` attribute.

        The operator is configured for future use. If an evolution
        operator was set earlier, it will be unset to maintain coherence.
        """

        num_vert = self._graph.number_of_vertices()
        num_arcs = self._graph.number_of_arcs()

        S_cols = [self._graph.arc_label(j, i)
                  for i in range(num_vert)
                  for j in self._graph.neighbors(i)]

        # Using csr_array((data, indices, indptr), shape)
        # Note that there is only one entry per row and column
        S = scipy.sparse.csr_array(
            ( np.ones(num_arcs, dtype=np.int8),
              S_cols, np.arange(num_arcs+1) ),
            shape=(num_arcs, num_arcs)
        )

        self._shift = S
        self._evolution = None

    def has_persistent_shift(self):
        r"""
        Checks whether the persistent shift operator is defined
        for the current graph and returns the result.

        The persistent shift operator can only be defined in a
        meaningful way for certain specific graphs. For instance,
        for graphs that can be embedded onto a plane,
        directions such as left, right, up, and down
        can be referred to naturally.
        """
        return self._graph.embeddable()

    def _set_persistent_shift(self):
        r"""
        Create the persistent shift operator (:math:`S`) based on
        the ``_graph`` attribute.

        The operator is set for future usage.
        If an evolution operator was set previously,
        it is unset for coherence.
        """
        num_arcs = self._graph.number_of_arcs()

        S_cols = [self._graph.previous_arc(i) for i in range(num_arcs)]

        # Using csr_array((data, indices, indptr), shape)
        # Note that there is only one entry per row and column
        S = scipy.sparse.csr_array(
            ( np.ones(num_arcs, dtype=np.int8),
              S_cols, np.arange(num_arcs+1) ),
            shape=(num_arcs, num_arcs)
        )

        self._shift = S
        self._evolution = None

    def set_shift(self, shift='default'):
        r"""
        Set the shift operator.

        Sets either the flipflop or the persistent shift operator.

        The generated shift operator is stored for future use in
        creating the evolution operator. If an evolution operator
        had been previously set, it is unset to maintain coherence.

        Parameters
        ----------
        shift: {'default', 'flipflop', 'persistent', 'ff', 'p'}
            Whether to create the flip flop or the persistent shift.
            By default, creates the persistent shift if it is defined;
            otherwise creates the flip flop shift.
            Argument ``'ff'`` is an alias for ``'flipflop'``.
            Argument ``'p'`` is an alias for ``'persistent'``.

        Raises
        ------
        AttributeError
            If ``shift='persistent'`` and
            the persistent shift operator is not implemented.

        See Also
        --------
        has_persistent_shift

        Notes
        -----
        .. note::
            Check :class:`CoinedWalk` Notes for details
            about the computational basis.


        The flip-flop shift operator :math:`S` is defined such that

        .. math::
            \begin{align*}
                S \ket{(v, u)} &= \ket{(u, v)} \\
                \implies S\ket i &= \ket j
            \end{align*}

        where :math:`i` is the label of the edge :math:`(v, u)` and
        :math:`j` is the label of the edge :math:`(u, v)`.


        For a more comprehensive understanding of the flipflop
        shift operator, refer to Section 7.2: Coined Walks on Arbitrary
        Graphs in the book "Quantum Walks and Search Algorithms" [1]_.
        

        References
        ----------
        .. [1] R. Portugal. "Quantum walks and search algorithms",
            2nd edition, Springer, 2018.

        Examples
        --------
        Consider the Graph presented in the
        :class:`Graph` Notes Section example.
        The corresponding flip-flop shift operator is

        .. testsetup::

            import numpy as np
            import scipy.sparse

        .. doctest::

            >>> import hiperwalk as hpw
            >>> A = scipy.sparse.csr_array([[0, 1, 0, 0],
            ...                             [1, 0, 1, 1],
            ...                             [0, 1, 0, 1],
            ...                             [0, 1, 1, 0]])
            >>> qw = hpw.CoinedWalk(A)
            >>> qw.set_shift(shift='flipflop')
            >>> S = qw.get_shift().todense()
            >>> S
            array([[0, 1, 0, 0, 0, 0, 0, 0],
                   [1, 0, 0, 0, 0, 0, 0, 0],
                   [0, 0, 0, 0, 1, 0, 0, 0],
                   [0, 0, 0, 0, 0, 0, 1, 0],
                   [0, 0, 1, 0, 0, 0, 0, 0],
                   [0, 0, 0, 0, 0, 0, 0, 1],
                   [0, 0, 0, 1, 0, 0, 0, 0],
                   [0, 0, 0, 0, 0, 1, 0, 0]], dtype=int8)

        Note that as required, :math:`S^2 = I`,
        :math:`S \ket 0 = \ket 1`, :math:`S \ket 1 = \ket 0`,
        :math:`S \ket 2 = \ket 4`, :math:`S \ket 4 = \ket 2`, etc.

        .. doctest::

            >>> (S @ S == np.eye(8)).all() # True by definition
            True
            >>> S @ np.array([1, 0, 0, 0, 0, 0, 0, 0]) # S|0> = |1>
            array([0, 1, 0, 0, 0, 0, 0, 0])
            >>> S @ np.array([0, 1, 0, 0, 0, 0, 0, 0]) # S|1> = |0>
            array([1, 0, 0, 0, 0, 0, 0, 0])
            >>> S @ np.array([0, 0, 1, 0, 0, 0, 0, 0]) # S|2> = |4>
            array([0, 0, 0, 0, 1, 0, 0, 0])
            >>> S @ np.array([0, 0, 0, 0, 1, 0, 0, 0]) # S|4> = |2>
            array([0, 0, 1, 0, 0, 0, 0, 0])

        .. todo::
            
            Add persistent example.
        """
        valid_keys = ['default', 'flipflop', 'persistent', 'ff', 'p']
        if shift not in valid_keys:
            raise ValueError(
                "Invalid `shift` value. Expected one of "
                + str(valid_keys) + ". But received '"
                + str(shift) + "' instead."
            )

        if shift == 'default':
            shift = 'p' if self.has_persistent_shift() else 'ff'

        if shift == 'ff':
            shift = 'flipflop'
        elif shift == 'p':
            shift = 'persistent'

        
        S = (self._set_flipflop_shift() if shift == 'flipflop'
             else self._set_persistent_shift())

        self._evolution = None

        if __DEBUG__:
            if self._shift is None: raise AssertionError

    def get_shift(self):
        return self._shift

    def set_coin(self, coin='default'):
        """
        Generate a coin operator based on the graph structure.

        Constructs a coin operator based on the degree of each vertex.
        A single type of coin may be applied to
        all vertices or a subset thereof.
        The coin operator is then established to be used in the generation
        of the evolution operator.

        Parameters
        ----------
        coin
            Coin to be used.
            Several types of arguments are acceptable.

            * str : coin type
                Type of the coin to be used.
                The following are valid entries.

                * 'default', 'd' : default coin,
                * 'fourier', 'F' : Fourier coin,
                * 'grover', 'G' : Grover coin,
                * 'hadamard', 'H' : Hadamard coin,
                * 'identity', 'I' : Identity,
                * 'minus_fourier', '-F' : Fourier coin with negative phase,
                * 'minus_grover', '-G' : Grover coin with negative phase,
                * 'minus_hadamard', '-H' : Hadamard coin with negative phase,
                * 'minus_identity', '-I' : Identity with negative phase.

            * list of str
                List of the coin types to be used.
                Expects list with 'number of vertices' entries.

            * dict
                A dictionary with structure
                ``{coin_type : list_of_vertices}``.
                That is, with any valid coin type as key and
                the list of vertices to be applied as values.
                If ``list_of_vertices = []``,
                the respective ``coin_type`` is applied to all vertices
                that were not explicitly listed.

            * :class:`scipy.sparse.csr_array`
                The explicit coin operator.

        Notes
        -----
        Owing to the selected computational basis (refer to the
        Notes in the :class:`CoinedWalk`),
        the outcome is a block diagonal operator.
        Each block is a :math:`\deg(v)`-dimensional ``coin``.
        As a result, there are :math:`|V|` blocks in total.

        .. todo::

            Check if explicit coin is valid.

        """
        try:
            if len(coin.shape) != 2:
                raise TypeError('Explicit coin is not a matrix.')

            # explicit coin
            if not scipy.sparse.issparse(coin):
                coin = scipy.sparse.csr_array(coin)

            self._coin = coin
            self._evolution = None
            return

        except AttributeError:
            pass

        coin_list, undefined_coin = self._coin_to_list(coin)
        if undefined_coin:
            raise ValueError('Coin was not specified for all vertices.')

        self._coin = coin_list
        self._evolution = None

        if __DEBUG__:
            if self._coin is None: raise AssertionError
            if self._evolution is not None: raise AssertionError

    def _coin_to_valid_name(self, coin):
        r"""
        Convert a string to its respective valid coin name.
        """
        s = coin
        if s == 'default' or s == 'd':
            s = self._graph.default_coin()
        
        if len(s) <= 2:
            prefix = 'minus_' if len(coin) == 2 else ''
            abbrv = {'F': 'fourier', 'G' : 'grover',
                     'H': 'hadamard', 'I': 'identity'}
            s = prefix + abbrv[s[-1]]

        if s not in CoinedWalk._coin_funcs.keys():
            raise ValueError(
                'Invalid coin. Expected any of '
                + str(list(CoinedWalk._coin_funcs.keys())) + ', '
                + "but received '" + str(coin) + "'."
            )

        return s

    def _coin_to_list(self, coin):
        r"""
        Convert str, list of str or dict to valid coin list.

        See Also
        --------
        set_coin
        """
        num_vert = self._graph.number_of_vertices()
        coin_list = []
        undefined_coin = False

        if isinstance(coin, str):
            coin_list = [self._coin_to_valid_name(coin)] * num_vert

        elif isinstance(coin, dict):
            coin_list = [''] * num_vert
            for key in coin:
                coin_name = self._coin_to_valid_name(key)
                value = coin[key]
                if value != []:
                    if not hasattr(value, '__iter__'):
                        value = [value]
                    for vertex in value:
                        coin_list[vertex] = coin_name
                else:
                    coin_list = [coin_name if coin_list[i] == ''
                                 else coin_list[i]
                                 for i in range(num_vert)]

            undefined_coin = '' in coin_list
        else:
            #list of coins
            if len(coin) != num_vert:
                raise ValueError('There were ' + str(len(coin))
                                 + ' coins specified. Expected '
                                 + str(num_vert) + 'coins instead.')

            coin_list = list(map(self._coin_to_valid_name, coin))

        return coin_list, undefined_coin

    @staticmethod
    def _fourier_coin(dim):
        return dft(dim, scale='sqrtn')

    @staticmethod
    def _grover_coin(dim):
        return np.array(2/dim * np.ones((dim, dim)) - np.identity(dim))

    @staticmethod
    def _hadamard_coin(dim):
        return hadamard(dim) / np.sqrt(dim)

    @staticmethod
    def _identity_coin(dim):
        return scipy.sparse.eye(dim)

    @staticmethod
    def _minus_fourier_coin(dim):
        return -CoinedWalk._fourier_coin(dim)

    @staticmethod
    def _minus_grover_coin(dim):
        return -CoinedWalk._grover_coin(dim)

    @staticmethod
    def _minus_hadamard_coin(dim):
        return -CoinedWalk._hadamard_coin(dim)

    @staticmethod
    def _minus_identity_coin(dim):
        return -np.identity(dim)

    def _coin_list_to_explicit_coin(self, coin_list):
        num_vert = self._graph.number_of_vertices()
        degree = self._graph.degree
        blocks = [self._coin_funcs[coin_list[v]](degree(v))
                  for v in range(num_vert)]
        C = scipy.sparse.block_diag(blocks, format='csr')
        return scipy.sparse.csr_array(C)

    #def get_coin(self):
    #    r"""
    #    Returns the coin operator in matricial form.
    #
    #    Returns
    #    -------
    #    :class:`scipy.sparse.csr_array`
    #    """
    #    if not scipy.sparse.issparse(self._coin):
    #        coin_list = self._coin
    #        C = self._coin_list_to_explicit_coin(coin_list)
    #
    #        if __DEBUG__:
    #            if not isinstance(C, scipy.sparse.csr_array):
    #                raise AssertionError
    #
    #        return C
    #    return self._coin

    def set_marked(self, marked=[]):
        r"""
        Set marked vertices.

        If a list of vertices is provided, those vertices are
        deemed as marked. However, the associated evolution operator
        remains unaffected.

        If a dictionary is passed,
        the coin of those vertices are substituted
        only for generating the evolution operator.
        This can only be done if the set coin operator is
        not a explicit matrix.

        If a dictionary is provided,
        the coins associated with those vertices  are replaced,
        but only for generating the evolution operator.
        Note that this can only be done if the coin operator
        is not set as an explicit matrix.

        Parameters
        ----------
        marked : list of int of dict
            list of vertices to be marked and
            how they are going to be marked.
            
            * list of int
                Given vertices are set as marked but
                the evolution operator remains unchanged.

            * dict
                A dictionary with structure
                ``{coin_type : list_of_vertices}``.
                Analogous to the one accepted by :meth:`set_coin`.

        See Also
        --------
        set_coin
        """
        coin_list = []

        if isinstance(marked, dict):
            coin_list, _ = self._coin_to_list(marked)

            dict_values = marked.values()
            vertices = [vlist if hasattr(vlist, '__iter__') else [vlist]
                        for vlist in dict_values]
            vertices = [v for vlist in vertices for v in vlist ]
            marked = vertices

        super().set_marked(marked)
        if bool(coin_list) or bool(self._oracle_coin):
            # evolution operator was changed
            self._evolution = None
        self._oracle_coin = coin_list

    #def get_oracle_coin(self):
    #    r"""
    #    The coins to be applied in the marked vertices.
    #
    #    Returns
    #    -------
    #    list of str
    #        If empty, no coin is changed (coin operator not altered).
    #        If an entry is the empty string,
    #        the coin for that vertex is not substituted.
    #    """
    #    return self._oracle_coin

    def get_coin(self):
        r"""
        Return coin to be used for creating the evolution operator.

        Returns
        -------
        :class:`scipy.sparse.csr_array`

        Notes
        -----
        The final coin :math:`C'` is obtained by multiplying the
        coin operator :math:`C` and the oracle :math:`R`.
        That is,

        .. math::
            
            C' = CR .

        The oracle is not explicitly saved.
        Instead, the oracle coins are saved -- i.e.
        which coin is going to be applied to each marked vertices.
        To generate :math:`C'` we simply substitute the original coin
        by the oracle coin in all marked vertices.

        Examples
        --------
        .. todo::
            examples
        """
        if scipy.sparse.issparse(self._coin):
            if not bool(self._oracle_coin):
                return self._coin

            # if coin was explicitly set,
            # and there are different coins for the marked vertices,
            # change them.
            def get_block(vertex):
                g = self._graph
                neighbors = g.neighbors(vertex)
                start = g.arc_label(vertex, neighbors[0])
                end = g.arc_label(vertex, neighbors[-1]) + 1

                return scipy.sparse.csr_array(self._coin[start:end,
                                                         start:end])

            num_vert = self.number_of_vertices()
            degree = self.degree
            oracle_coin = self._oracle_coin
            coin_funcs = CoinedWalk._coin_funcs
            blocks = [coin_funcs[oracle_coin[v]](degree(v))
                      if oracle_coin[v] != ''
                      else get_block(v)
                      for v in range(num_vert)]
            C = scipy.sparse.block_diag(blocks, format='csr')

            return scipy.sparse.csr_array(C)

        oracle_coin = self._oracle_coin
        if bool(oracle_coin):
            coin = self._coin
            coin_list = [oracle_coin[i] if oracle_coin[i] != ''
                         else coin[i]
                         for i in range(len(coin))]
        else:
            coin_list = self._coin

        return self._coin_list_to_explicit_coin(coin_list)

    def set_evolution(self, **kwargs):
        """
        Set the evolution operator.

        Shorthand for setting shift, coin and marked vertices.
        They are set using the appropriate ``**kwargs``.
        If ``**kwargs`` is empty, the default arguments are used.

        Parameters
        ----------
        **kwargs : dict, optional
            Arguments for setting the evolution operator.
            Accepts any valid keywords from
            :meth:`set_shift` :meth:`set_coin`, and :meth:`set_marked`.

        See Also
        --------
        set_shift
        set_coin
        set_marked
        """

        S_kwargs = CoinedWalk._filter_valid_kwargs(
                              kwargs,
                              CoinedWalk._valid_kwargs['shift'])
        C_kwargs = CoinedWalk._filter_valid_kwargs(
                              kwargs,
                              CoinedWalk._valid_kwargs['coin'])
        R_kwargs = CoinedWalk._filter_valid_kwargs(
                              kwargs,
                              CoinedWalk._valid_kwargs['marked'])

        self.set_shift(**S_kwargs)
        self.set_coin(**C_kwargs)
        self.set_marked(**R_kwargs)
        self._evolution = None

    def get_evolution(self, hpc=True):
        r"""
        Create evolution operator from previously set attributes.

        Creates evolution operator by multiplying the
        shift operator and coin operator.
        If the coin operator is not an explicit matrix,
        and the coin for marked vertices was specified,
        the coin of each marked vertex is substituted.

        Parameters
        ----------
        hpc : bool, default=True
            Whether or not the evolution operator should be
            constructed using nelina's high-performance computing.

        Returns
        -------
        :class:`scipy.sparse.csr_array`

        Notes
        -----
        The evolution operator is given by

        .. math::
           U = SC'

        where :math`S` is the shift operator, and
        :math:`C'` is the coin operator (probably) modified by
        the marked vertices [1]_.

        If the coin operator was set as an explicit matrix,
        the marked vertices to not alter it.
        If the coin operator was not set as an explicit matrix
        (e.g. as a list of coins),
        the coin of each marked vertex is substituted as specified by
        the last :meth:`set_marked` call.

        .. todo::
            * Sparse matrix multipliation is not supported yet.
              Converting all matrices to dense.
              Then converting back to sparse.
              This uses unnecessary memory and computational time.
            * Check if matrix is sparse in pynelibna interface
            * Check if matrices are deleted from memory and GPU.


        References
        ----------
        .. [1] R. Portugal. "Quantum walks and search algorithms",
            2nd edition, Springer, 2018.

        Examples
        --------

        .. todo::
            Valid examples to clear behavior.
        """
        if self._evolution is not None:
            # evolution operator was not changed.
            # No need to create it again
            return self._evolution

        U = None
        if hpc and not self._pyneblina_imported():
            hpc = False


        S = self.get_shift()
        C = self.get_coin()

        if hpc:

            S = S.todense()
            C = C.todense()

            nbl_S = nbl.send_matrix(S)
            del S
            nbl_C = nbl.send_matrix(C)
            del C
            nbl_C = nbl.multiply_matrices(nbl_S, nbl_C)

            del nbl_S

            U = nbl.retrieve_matrix(nbl_C)
            del nbl_C
            U = scipy.sparse.csr_array(U)

        else:
            U = S @ C

        self._evolution = U
        return U


    def probability_distribution(self, states):
        """
        Compute the probability distribution of the given states.

        The probability of finding the walker at each vertex
        for the given states.

        Parameters
        ----------
        states : :class:`numpy.ndarray`
            The states used to compute the probabilities.

        Returns
        -------
        probabilities : :class:`numpy.ndarray`

        See Also
        --------
        simulate

        Notes
        -----
        .. note::
            
            benchmark performance
        """
        # TODO: test with nonregular graph
        # TODO: test with nonuniform condition
        if __DEBUG__:
            start = now()

        if len(states.shape) == 1:
            states = [states]

        #edges_indices = self.adj_matrix.indptr
        #
        #prob = np.array([[
        #        Graph._elementwise_probability(
        #            states[i][edges_indices[j]:edges_indices[j + 1]]
        #        ).sum()
        #        for j in range(len(edges_indices) - 1)
        #    ] for i in range(len(states)) ])

        def get_entries(state, indexes):
            return np.array([state[i] for i in indexes])

        num_vert = self._graph.number_of_vertices()
        graph = self._graph
        prob = np.array([[CoinedWalk._elementwise_probability(
                              get_entries(
                                  states[i],
                                  graph.arcs_with_tail(v)
                              )
                          ).sum()
                          for v in range(num_vert)]
                        for i in range(len(states))])

        # TODO: benchmark (time and memory usage)
        return prob

    def state(self, *args):
        """
        Generates a valid state.

        The state corresponds to the walker being in a superposition
        of the ``entries``.
        For instance, click on :meth:`qwalk.coined.Graph`.
        The final state is normalized in order to be a unit vector.

        Parameters
        ----------
        *args
            Each entry is a tuple (or array).
            An entry can be specified in three different ways:
            ``(amplitude, (vertex, dst_vertex))``,
            ``(amplitude, arc_label)``,
            ``(amplitude, vertex, coin)``.

            amplitude
                The amplitude of the given entry.
            vertex
                The vertex corresponding to the position of the walker
                in the superposition.
            dst_vertex
                The vertex to which the coin is pointing.
                In other words, the tuple
                ``(vertex, dst_vertex)`` must be a valid arc.
            arc_label
                The arc label with respect to the arc ordering
                given by the computational basis.
            coin
                The direction towards which the coin is pointing.
                It is dependable on the Graph coloring.

        Notes
        -----
        If entries are repeated (except by the amplitude),
        they are overwritten by the last one.

        Examples
        --------
        .. todo::
            
            qw.state([1/np.sqrt(2), 0], [1/np.sqrt(2), (1, 0)])
        """

        has_complex = np.any([arg[0].imag != 0 for arg in args])
        state = np.zeros(self.hilb_dim, dtype=complex if has_complex
                                                      else float)

        for entry in args:
            if len(entry) == 3:
                raise NotImplementedError(
                    'position-coin notation not implemented')

            if hasattr(entry[1], '__iter__'):
                # arc notation
                head, tail = entry[1]
                state[self._graph.arc_label(head, tail)] = entry[0]
            else:
                # arc label
                state[entry[1]] = entry[0]

        return self._normalize(state)

    def ket(self, *args):
        r"""
        Create a computational basis state.

        Parameters
        ----------
        *args
            The ket label.
            There are two different labels acceptable.

            tail, head
                The arc notation.
            arc_label
                The label of the arc.
                Its number according to the computational basis order.

        Examples
        --------
        .. todo::
            valid examples
        """
        ket = np.zeros(self.hilb_dim, dtype=float)
        if len(args) == 2:
            ket[self._graph.arc_label(args[0], args[1])] = 1
        else:
            ket[args] = 1

        return ket



import hiperwalk as hpw

dim = 121
lat = hpw.Lattice(dim, diagonal=True)
center = lat.get_central_vertex()
dtqw = hpw.CoinedWalk(lat, shift='persistent', coin='grover')
psi0 = dtqw.state([0.5, (center, center + (1, 1))],
                  [-0.5, (center, center + (1, -1))],
                  [-0.5, (center, center + (-1, 1))],
                  [0.5, (center, center + (-1, -1))])
psi_final = dtqw.simulate(time=dim // 2, initial_state=psi0, hpc=False)
prob = dtqw.probability_distribution(psi_final)
hpw.plot_probability_distribution(prob, graph=lat)




import numpy as np
import networkx as nx
import random
from sys import path as sys_path
sys_path.append('../')
sys_path.append('../../')
from test_constants import *
import hiperwalk as hpw
import unittest

class TestContinuousGraph(unittest.TestCase):

    def setUp(self):
        self.num_vert = 40
        nx_graph = nx.ladder_graph(self.num_vert // 2)
        self.adj = nx.adjacency_matrix(nx_graph)
        self.g = hpw.Graph(self.adj)
        self.gamma = 1/2
        self.qw = hpw.ContinuousWalk(self.g, gamma=self.gamma)

    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_hamiltonian_default(self):
        H = self.qw.get_hamiltonian()
        self.assertTrue((H - (-self.gamma*self.adj) != 0).nnz == 0)
        self.assertTrue(self.qw._hamiltonian is not None)
        self.assertTrue(self.qw._marked == [])

    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_hamiltonian_multiple_marked(self):
        prev_H = self.qw._hamiltonian
        prev_marked = self.qw._marked

        gamma = 1/3
        num_marked = random.randint(0, self.num_vert - 1)
        marked = random.sample(list(range(self.num_vert)), num_marked)
        self.qw.set_hamiltonian(gamma=gamma, marked=marked)

        H = self.qw.get_hamiltonian()

        marked_matrix = np.diag([1 if i in marked else 0
                                 for i in range(self.num_vert)])
        self.assertTrue(np.all(H - (-gamma*self.adj - marked_matrix) == 0))
        self.assertTrue(self.qw._hamiltonian is not None)
        self.assertTrue(id(prev_H) != id(self.qw._hamiltonian))
        self.assertTrue(self.qw._marked is not None)
        self.assertTrue(id(self.qw._marked) != id(prev_marked))
        
    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_evolution_operator_invalid_time(self):
        marked = self.qw._marked
        H = self.qw._hamiltonian
        U = self.qw._evolution

        self.assertRaises(ValueError, self.qw.get_evolution,
                          time=-1, hpc=False)

        self.assertTrue(id(marked) == id(self.qw._marked))
        self.assertTrue(id(H) == id(self.qw._hamiltonian))
        self.assertTrue(id(U) == id(self.qw._evolution))

    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_evolution_operator_set_hamiltonian_no_marked(self):
        prev_marked = self.qw._marked
        prev_H = self.qw._hamiltonian
        prev_U = self.qw._evolution

        self.qw.set_evolution(gamma=1, marked=[])
        U = self.qw.get_evolution(time=1, hpc=False)

        self.assertTrue(U is not None)
        self.assertTrue(self.qw._evolution is not None)
        self.assertTrue(id(prev_marked) != id(self.qw._marked))
        self.assertTrue(id(prev_H) != id(self.qw._hamiltonian))
        self.assertTrue(id(prev_U) != id(self.qw._evolution))

    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_evolution_operator_set_hamiltonian_and_marked(self):
        prev_marked = self.qw._marked
        prev_H = self.qw._hamiltonian
        prev_U = self.qw._evolution

        self.qw.set_evolution(gamma=1, marked=0)
        U = self.qw.get_evolution(time=1, hpc=False)

        self.assertTrue(U is not None)
        self.assertTrue(self.qw._evolution is not None)
        self.assertTrue(id(prev_marked) != id(self.qw._marked))
        self.assertTrue(id(prev_H) != id(self.qw._hamiltonian))
        self.assertTrue(id(prev_U) != id(self.qw._evolution))

    @unittest.skipIf(not TEST_NONHPC, 'Skipping nonhpc tests.')
    def test_evolution_unitary(self):
        U = self.qw.get_evolution(time=1, hpc=False)

        self.assertTrue(np.allclose(
            U@U.T.conjugate(), np.eye(U.shape[0])
        ))

    @unittest.skipIf(not TEST_HPC, 'Skipping hpc tests.')
    def test_hpc_evolution_unitary(self):
        U = self.qw.get_evolution(time=1, hpc=True)

        self.assertTrue(np.allclose(
            U@U.T.conjugate(), np.eye(U.shape[0])
        ))



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 14 16:04:28 2017

@author: tharso
"""
import config as cfg
import networkx as nx
import os


def read_adj_matrix(file_name):
    G = nx.Graph()

    with open(file_name) as f_r:
        content = f_r.readlines()

    for k in range(0, len(content)):
        inf = content[k].split()
        if int(inf[2]) != 0:
            #print [int(inf[0]),int(inf[1])]
            G.add_edges_from([(int(inf[0]), int(inf[1]))])

    nx.draw(G)
    return G


def dijkstra(ProbPsi, PsiSize, G, file_name):
    print("[Hiperwalk] Computing distances...")
    alldists = nx.all_pairs_shortest_path_length(G)
    f = open(file_name, 'w')
    d = [0.] * PsiSize
    for i in ProbPsi:
        index = i[0]
        prob = i[1]
        for j in range(PsiSize):
            d[j] += prob * alldists[index][j + 1]
    for k in range(len(d)):
        f.write(str(d[k]) + '\n')
    f.close()


def nonblank_lines(f):
    for l in f:
        line = l.rstrip()
        if line:
            yield line


def generateDistance():
    fn = cfg.ADJMATRIX_PATH
    G = read_adj_matrix(fn)
    curpath = os.path.abspath(os.curdir)
    psiFileName = cfg.CUSTOM_INITIALSTATE_NAME
    linecount = 1
    PsiProb = []
    with open(psiFileName) as f_in:
        for line in nonblank_lines(f_in):
            num = line.split(' ')
            if len(num) == 1:
                re = float(num[0])
                im = 0.0
            elif len(num) >= 2:
                re = float(num[0])
                im = float(num[1])

            if (re != 0. or im != 0.):
                PsiProb.append([linecount, re * re + im * im])
            linecount += 1
    dijkstra(PsiProb, linecount - 1, G, curpath + "/distance_custom.dat")



# -*- coding: utf-8 -*-
"""

Created on Mon Oct 27 04:06:41 2014

@author: aaron
"""
import neblina as nb
import testmode
import config as cfg
import numpy as np

def run():
    print("[Hiperwalk] CUSTOM WALK.")
    nb.runCore_CUSTOM()

    
    if cfg.TEST_MODE:
        modelVector=testmode.create_CUSTOM_test_vector()
        returnNeblina=nb.neblina_distribution_to_vector("final_distribution.dat")
        if np.linalg.norm(modelVector-returnNeblina,np.inf) == float(0):
            return 1
        else:
            return 0
            
            
    return 1


# -*- coding: utf-8 -*-
"""
Created on Sun Sep 14 16:19:56 2014

@author: aaron
"""
import config as cfg
import ioFunctions as io
#import numpy as np
import gnuplot as gnuplot
import neblina as nb
import operators as op
import standardDeviation as sd
import testmode
import numpy as np

def run():
    probabilities=[]

    io.savetxt("HIPERWALK_TEMP_PSI.dat",cfg.STATE,float,'%1.16f')
    op.COIN_TENSOR_IDENTITY_1D(cfg.COINOPERATOR,cfg.GRAPHSIZE)
    op.SHIFT_OPERATOR_1D(cfg.COINVECTORDIMENSION,cfg.GRAPHSIZE)
#    sd.distances_vector_1D(cfg.RANGEX[0],cfg.RANGEX[1])  
    cfg.DISTANCE_VECTOR_SIZE=cfg.GRAPHSIZE
#    io.savetxt("HIPERWALK_TEMP_DISTANCE_VECTOR.dat",cfg.DISTANCES_VECTOR,int,'%1.16f')    

#    nb.generating_DTQW1D_NBL()
    nb.runCore_DTQW1D()

    cfg.STATE=nb.neblina_state_to_vector("NEBLINA_TEMP_final_state.dat")
    probabilities=nb.neblina_distribution_to_vector("NEBLINA_TEMP_final_distribution.dat")
    output = open("final_distribution.dat",'w')
    output1 = open("final_state.dat",'w')
    output.write("#POSITION \t PROBABILITY\n")          
    output1.write("#POSITION \t COIN \t Re(amplitude) \t Im(amplitude)\n")        
    
    for j in range(cfg.COINVECTORDIMENSION):
        for i in range(int(cfg.GRAPHSIZE)):
            if j==0:
                output.write("%d \t \t \t%1.16f\n"%(cfg.RANGEX[0]+i,probabilities[i]))
            output1.write("%d \t \t \t \t %d\t \t %1.16f \t %1.16f\n"%(cfg.RANGEX[0]+i,j,cfg.STATE[i+j*cfg.GRAPHSIZE].real,cfg.STATE[i+j*cfg.GRAPHSIZE].imag))         

    output.close()
    output1.close()    
            

        





    if cfg.GNUPLOT:
        io.savetxt("HIPERWALK_TEMP_PROBABILITIES.dat",probabilities,float,'%1.16f')
        gnuplot.template_DTQW1D("HIPERWALK_TEMP_PROBABILITIES.dat","final_distribution.eps","EPS")
        if cfg.STEPS>1:
            gnuplot.plotStatistics1D()
        if cfg.ANIMATION == 1:
            gnuplot.plotAnimation1D()
        



    if cfg.TEST_MODE:
        modelVector=testmode.create_DTQW1D_test_vector()
        returnNeblina=nb.neblina_distribution_to_vector("NEBLINA_TEMP_final_distribution.dat")
        if np.linalg.norm(modelVector-returnNeblina,np.inf) == float(0):
            return 1
        else:
            return 0
            
    return 1


import networkx as nx #TODO: import only needed functions?
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
from ._animation import *
from .._constants import __DEBUG__
from ..graph import *

if __DEBUG__:
    from time import time

# TODO: move to constants
plt.rcParams["figure.figsize"] = (12, 10)
plt.rcParams["figure.dpi"] = 100


# TODO: add option for changing figsize and dpi
# histogram is alias for bar width=1
def plot_probability_distribution(
        probabilities, plot=None, animate=False, show=True,
        filename=None, interval=250, **kwargs):
    """
    Plots probability distribution of quantum walk.

    The probability distribution of multiple steps of a quantum walk
    may be plotted.
    The generated figures may be shown step-by-step,
    saved in multiple files, animated or saved as animation.

    Parameters
    ----------
    probabilities : :class:`numpy.ndarray`
        The probabilities of the walker to be found on each step
        of the quantum walk.
        Columns represent vertices and rows represent the walk steps.
    plot : str, default=None
        The plot type.
        The valid options are
        ``{'bar', 'line', 'graph', 'histogram', 'plane'}``.
        If ``None``, uses default plotting. Usually ``bar``,
        but default plotting changes according to ``graph``.
    show : bool, default=True
        Whether or not to show plots or animation.
        With ``show=True`` we have:
        **Using Terminal**:
        After shown, press *q* to quit.
        If ``animate==False``,
        the quantum walk will be shown step-by-step;
        If ``animate==True``,
        the entire animation is shown in a new window.
        **Using Jupyter Notebook**:
        If ``animate==False``,
        all the quantum walk steps are shown at once.
        If ``animate==True``,
        the entire animation is shown as a html video.
    filename : str, default=None
        The filename path (with no format) where
        the plot(s) will be saved.
        If ``None`` no file is saved.
        Otherwise, if ``animate==False``,
        the j-step is saved in the ``filename-j.png`` file;
        if ``animate==True``,
        the entire walk is saved in the ``filename.fig`` file.
    graph : optional
        The structure of the graph on which the walk occurs.
        The graph labels are used as plotting labels.
        **Important**: check Graph Plots subsection in other parameters.

        The following types are acceptable.
        
        * :class:`hiperwalk.Graph`
            Hiperwalk Graph.
            It is used to generate default plotting for specific graphs.
            Does not override values explicited by the user.
        * :class:`networkx.classes.graph`,
            NetworkX Graph
        * :class:`scipy.sparse.csr_matrix`
            Adjacency matrix.
    rescale : bool, optional
        If ``False`` or omitted, the reference maximum probability
        is the global one.
        If ``True``, the reference maximum probability depends on
        the current step, changing every image or frame.
        For example, if the global maximum probability is 1,
        ``min_node_size, max_node_size = (300, 3000)``,
        and the maximum probability of a given step is 0.5;
        then for ``rescale=False``,
        the step maximum node size shown is 1650
        (halfway betweeen 300 and 3000),
        while for ``rescale=True``,
        the step maximum node size shown is 3000.
    animate : bool, default=False
        Whether or not to animate multiple plots.
        If ``False``, each quantum walk step generates an image.
        If ``True``, each quantum walk step is used as an animation frame.
    interval : int, default=250
        Time in milliseconds that each frame is shown if ``animate==True``.
    **kwargs : dict, optional
        Extra arguments to further customize plotting.
        Valid arguments depend on ``plot``.
        Check Other Parameters Section for details.

    Other Parameters
    ----------------
    Bar Plots
        See :obj:`matplotlib.pyplot.bar` for more optional keywords.

    Graph Plots
        See :obj:`networkx.draw <networkx.drawing.nx_pylab.draw>`
        for more optional keywords.

        graph :
            Graph structure.
        min_node_size, max_node_size : scalar, default=(300, 3000)
            By default, nodes sizes depend on the probability.
            ``min_node_size`` and ``max_node_size`` describe the
            inferior and superior limits for the node sizes, respectively.
        node_size : scalar or list of scalars, optional
            If ``node_size`` is a scalar,
            all nodes will have the same size.
            If ``node_size`` is a list of scalars,
            vertices may have different sizes and
            the length of ``node_size`` must match ``probabilities``.
            The ``node_size`` argument is ignored if both
            ``min_node_size`` and ``max_node_size`` are set.
        cmap : str, optional
            A colormap for representing vertices probabilities.
            if ``cmap='default'``, uses the ``'viridis'`` colormap.
            For more colormap options, check
            `Matplolib's Colormap reference
            <https://matplotlib.org/stable/gallery/color/colormap_reference.html>`_.

    Histogram Plots
        See :obj:`matplotlib.pyplot.bar` for more optional keywords.
        The ``width`` keyword is always overriden.

    Line Plots
        See :obj:`matplotlib.pyplot.plot` for more optional keywords.

    Plane Plots
        dimensions: 2-tuple of int
            plane dimensions in ``(x_dim, y_dim)`` format.


    Raises
    ------
    ValueError
        If ``plot`` has an invalid value.
    KeyError
        If ``plot == 'graph' `` and keyword ``graph`` is not set.

    Warnings
    --------
    For showing animations,
    the current version only supports Jupyter and GTK 3.0.
    It must be updated to support GTK 4.0 for Ubuntu 22.04.

    Notes
    -----
    The core logic of the main implementation loop is more or less like follows.

    >>> preconfigure() # doctest: +SKIP
    >>> for prob in probabilities: # doctest: +SKIP
    >>>     configure() # doctest: +SKIP
    >>>     plot(prob) # doctest: +SKIP

    ``preconfigure()`` executes configurations that do
    not change between plots, e.g. nodes positions in graph plots.
    ``configure()`` executes configurates that must be
    (re)done for each iteration, e.g. setting figure size.
    ``plot()`` calls the appropriated plotting method with
    customization parameters,
    i.e. bar, line or graph plot with the respective valid kwargs.

    .. todo::
        - Accept the ``probabilities`` parameter as a list.
        - Use Graphviz instead of NetworkX to draw graphs.
            As noted by networkx's documentation:
            Proper graph visualization is hard, and we highly recommend
            that people visualize their graphs with tools dedicated to
            that task. 
            https://networkx.org/documentation/stable/reference/drawing.html
        - Implement ``repeat_delay`` parameter:
            An extra time to wait before the animation is repeated.
            Pull requests to the
            `Matplotlib animation writers
            <https://matplotlib.org/stable/api/animation_api.html#writer-classes>`_
            are needed.
        - Implement ``transparency`` parameter:
            change nodes transparency depending on probability.
        - Implement GTK 4.0 support.

    Examples
    --------
    .. todo::
        probabilities expects numpy array or matrix
    """
    # passes kwargs by reference to be updated accordingly

    if 'graph' in kwargs:
        plot = _default_graph_kwargs(kwargs, plot)

    if plot is None:
        plot = 'bar'

    plot = plot.lower()
    valid_plots = ['bar', 'line', 'graph', 'histogram', 'plane']

    if plot not in valid_plots:
        raise ValueError(
            'Unexpected value for plot:' + str(plot) +
            '. One of the following was expected: ' + str(valid_plots)
        )

    # dictionaries for function pointers
    # preconfiguration: executed once before the loop starts
    preconfigs = {valid_plots[0]: _preconfigure_plot,
            valid_plots[1]: _preconfigure_plot,
            valid_plots[2]: _preconfigure_graph_plot,
            valid_plots[3]: _preconfigure_plot,
            valid_plots[4]: _preconfigure_plot}
    # configuration: executed every iteration before plotting
    # expects return of fig, ax to be used for animations
    configs = {valid_plots[0]: _configure_plot_figure,
            valid_plots[1]: _configure_plot_figure,
            valid_plots[2]: _configure_graph_figure,
            valid_plots[3]: _configure_plot_figure,
            valid_plots[4]: _configure_plane_figure}
    # plot functions: code for plotting the graph accordingly
    plot_funcs = {
        valid_plots[0]: _plot_probability_distribution_on_bars,
        valid_plots[1]: _plot_probability_distribution_on_line,
        valid_plots[2]: _plot_probability_distribution_on_graph,
        valid_plots[3]: _plot_probability_distribution_on_histogram,
        valid_plots[4]: _plot_probability_distribution_on_plane
    }

    # preparing probabilities to shape requested by called functions
    if len(probabilities.shape) == 1:
        probabilities = [probabilities]

    # passes kwargs by reference to be updated accordingly
    preconfigs[plot](probabilities, kwargs)

    if animate:
        anim = Animation()

    for i in range(len(probabilities)):
        # TODO: set figure size according to graph dimensions
        # TODO: check for kwargs
        fig, ax = configs[plot](probabilities.shape[1]) 

        plot_funcs[plot](probabilities[i], ax, **kwargs)

        plt.tight_layout()

        # saves or shows image (or both)
        if not animate:
            if filename is not None:
                # enumarating the plot
                filename_suffix = (
                    '-' + (len(probabilities)-1)//10 * '0' + str(i)
                    if len(probabilities) > 1 else ''
                )
                plt.savefig(filename + filename_suffix)
                if not show:
                    plt.close()
            if show:
                plt.show()

        else:
            anim.add_frame(fig)

    if animate:
        anim.create_animation(interval)

        if filename is not None:
            anim.save_animation(filename)
        if show:
            anim.show_animation()

def _default_graph_kwargs(kwargs, plot):
    if ((plot is None or plot == 'graph' or plot == 'plane')
        and not 'cmap' in kwargs
    ):
        kwargs['cmap'] = 'default'

    if 'cmap' in kwargs:
        if kwargs['cmap'] == 'default':
            kwargs['cmap'] = 'viridis'

    graph = kwargs['graph']

    # hiperwalk graph
    if isinstance(graph, Lattice):
        if plot is None:
            plot = 'plane'
        if plot == 'plane' and 'dimensions' not in kwargs:
            kwargs['dimensions'] = graph.dimensions()
        return plot

    if isinstance(graph, Hypercube):
        if plot is None:
            plot = 'graph'

            nx_graph = nx.from_scipy_sparse_array(graph.adj_matrix)
            for v in nx_graph:
                try:
                    nx_graph.nodes[v]["subset"] = v.bit_count()
                except AttributeError:
                    nx_graph.nodes[v]["subset"] = bin(v).count('1')

            kwargs['pos'] = nx.multipartite_layout(nx_graph)
            dim = graph.dimension()
            kwargs['graph'] = nx_graph
            kwargs['edge_color'] = (0, 0, 0, max(0.002, 2**(-dim/2)))
            kwargs['labels'] = {0: 0, 2**dim - 1 : 2**dim - 1}
            kwargs['min_node_size'] = 1
            kwargs['max_node_size'] = 1000
            kwargs['edgecolors'] = None

    return 'graph' if plot is None else plot


def _preconfigure_plot(probabilities, kwargs):
    """
    Configure static parameters for matplotlib plot.

    Set parameters that need not to be changed for multiple plots
    or animation of a quantum walk.
    For example: minimum and maximum allowed node size.

    Parameters
    ----------
    probabilities : list of floats
        Probabilities of the walker being found on each vertex.
    kwargs : dict
        Reference of kwargs containing all extra keywords.
    """

    if ('rescale' not in kwargs or not kwargs.pop('rescale')):
        kwargs['min_prob'] = 0
        kwargs['max_prob'] = probabilities.max()


#kwargs passed by reference
def _preconfigure_graph_plot(probabilities, kwargs):
    """
    Configure static parameters for graph plot.

    Set parameters that need not to be changed for multiple plots
    or animation of a quantum walk.
    For example: the graph.
    The graph must be sent via kwargs,
    either via the 'graph' or the 'adj_matrix' keywords.

    Parameters
    ----------
    probabilities : list of floats
        Probabilities of the walker being found on each vertex.
    kwargs : dict
        Reference of kwargs containing all extra keywords.

    See Also
    --------
    _configure_nodes
    """

    if ('rescale' not in kwargs or not kwargs['rescale']):
        kwargs['min_prob'] = 0 #min_prob
        kwargs['max_prob'] = probabilities.max() #max_prob
        kwargs['rescale'] = False

    if 'graph' not in kwargs:
        raise KeyError("'graph' kwarg not provided.")

    graph = kwargs['graph']
    if isinstance(graph, scipy.sparse.csr_array):
        kwargs['graph'] = nx.from_scipy_sparse_array(graph)
    elif isinstance(graph, Graph):
        kwargs['graph'] = nx.from_scipy_sparse_array(graph.adj_matrix)

    if 'adj_matrix' in kwargs:
        # Pops adj_matrix if both graph and adj_matrix keywords are set;
        # otherwise invalid keyword may be raisen by networkx
        kwargs.pop('adj_matrix')

    if 'min_node_size' not in kwargs:
        kwargs['min_node_size'] = None
    if 'max_node_size' not in kwargs:
        kwargs['max_node_size'] = None

    # setting static kwargs for plotting
    # kwargs dictionary is updated by reference
    # TODO: change ConfigureNodes parameters
    # (remove G and use information from kwargs)
    _configure_nodes(kwargs['graph'], probabilities, kwargs)


def _configure_figure(num_vert, fig_width=None, fig_height=None):
    """
    Set basic figure configuration.

    Creates a figure with size depending on the parameters.

    Parameters
    ----------
    num_vert: int
        number of vertices in the graph
        .. todo:: set the figure size according to num_vert
    fig_width, fig_height : float, optional
        Custom figure width and height, respectively.
        If not set, the default value is used.

    Returns
    -------
    fig : current figure
    ax : current figure axes
    """

    #TODO: set figure size according to graph dimensions
    if fig_width is None:
        fig_width = plt.rcParams["figure.figsize"][0]
    if fig_height is None:
        fig_height = plt.rcParams["figure.figsize"][1]

    fig = plt.figure(figsize=(fig_width, fig_height))
    ax = plt.gca()

    return fig, ax


def _configure_plot_figure(num_vert, fig_width=None, fig_height=None):
    """
    Set basic figure configuration for matplotlib plots.
    """
    
    fig, ax = _configure_figure(num_vert, fig_width, fig_height)

    plt.xlabel("Vertex", size=18)
    plt.ylabel("Probability", size=18)

    plt.tick_params(length=7, labelsize=14)

    return fig, ax


def _configure_graph_figure(num_vert=None, fig_width=None,
                            fig_height=None):
    return _configure_figure(num_vert, fig_width, fig_height)

def _configure_plane_figure(num_vert=None, fig_width=None,
                           fig_height=None):
    if fig_width is None:
        fig_width = plt.rcParams["figure.figsize"][0]
    if fig_height is None:
        fig_height = plt.rcParams["figure.figsize"][1]

    #fig, ax =_configure_figure(num_vert, fig_width, fig_height) 
    #ax = fig.add_subplot(projection='3d')
    fig, ax = plt.subplots(figsize=(fig_width, fig_height),
                           subplot_kw={"projection": "3d"})

    ax.tick_params(length=10, width=1, labelsize=16, pad=10)
    ax.set_xlabel('Vertex X', labelpad=15, fontsize=18)
    ax.set_ylabel('Vertex Y', labelpad=15, fontsize=18)
    ax.set_zlabel('Probability', labelpad=30, fontsize=18)
    return fig, ax


def _plot_probability_distribution_on_bars(
        probabilities, ax, labels=None, graph=None,
        min_prob=None, max_prob=None, **kwargs
    ):
    """
    Plot probability distribution using matplotlib bar plot.

    Parameters
    ----------
    probabilities : list of floats
        Probabilities of the walker being found on each vertex.
    ax
        matplotlib ax on which the figure is drawn.
    {labels, graph, min_prob, max_prob} : optional
        Final configuration parameters.
        Refer to _posconfigure_plot_figure.
    **kwargs : dict, optional
        Extra parameters for plotting. Refer to matplotlib.pyplot.bar

    See Also
    --------
    _posconfigure_plot_figure : Sets final configuraation for exhibition.
    matplotlib.pyplot.bar
    """

    plt.bar(np.arange(len(probabilities)), probabilities, **kwargs)
    _posconfigure_plot_figure(ax, len(probabilities), labels, graph,
                             min_prob, max_prob)


def _plot_probability_distribution_on_histogram(
        probabilities, ax, labels=None, graph=None,
        min_prob=None, max_prob=None, **kwargs
    ):
    """
    Plot probability distribution as histogram.

    Alias for no space between bars in bar plot.

    Parameters
    ----------
    Refer to _plot_probability_distribution_on_bars

    See Also
    --------
    _plot_probability_distribution_on_bars
    """
    
    kwargs['width'] = 1
    _plot_probability_distribution_on_bars(
        probabilities, ax, labels, graph, min_prob, max_prob, **kwargs
    )


def _plot_probability_distribution_on_line(
        probabilities, ax, labels=None, graph=None,
        min_prob=None, max_prob=None, **kwargs
    ):
    """
    Plots probability distribution using matplotlib's line plot.

    Parameters
    ----------
    probabilities : list of floats
        Probabilities of the walker being found on each vertex.
    ax
        matplotlib ax on which the figure is drawn.
    {labels, graph, min_prob, max_prob} : optional
        Final configuration parameters.
        Refer to _posconfigure_plot_figure.
    **kwargs : dict, optional
        Extra parameters for plotting. Refer to matplotlib.pyplot.plot

    See Also
    --------
    _posconfigure_plot_figure : Sets final configuraation for exhibition.
    matplotlib.pyplot.plot
    """

    if 'marker' not in kwargs:
        kwargs['marker'] = 'o'
    plt.plot(np.arange(len(probabilities)), probabilities, **kwargs)

    _posconfigure_plot_figure(
        ax, len(probabilities), labels, graph, min_prob, max_prob
    )

def _posconfigure_plot_figure(ax, num_vert, labels=None, graph=None,
                              min_prob=None, max_prob=None):
    """
    Add final touches to the plotted figure.

    The labels are added to the figure and
    the figure's y-axis range is configured.

    Parameters
    ----------
    ax
        matplotlib ax on which the figure is drawn
    num_vert : int
        number of vertices in the graph
    labels : dict, optional
        Labels to be shown on graph.
        If `graph` is `None`, `labels.keys()` must be integers
        from 0 to `num_vert` - 1.
    graph : networkx graph, optional
        Graph on which the quantum walk occured.
        if not None, the graph nodes are used to match `labels` keys.
    min_prob, max_prob : int, optional
        If both are set, describe the y-axis limit.
    """
    if labels is not None:
        if graph is None:
            ax.set_xticks( list(labels.keys()), list(labels.values()) )
        else:

            nodes = list(graph.nodes())
            nodes = {i : labels[ nodes[i] ] for i in range(num_vert)
                        if nodes[i] in labels}

            ax.set_xticks( list(nodes.keys()), list(nodes.values()) )

    else:
        from matplotlib.ticker import MaxNLocator

        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        if graph is not None:

            ax.set_xlim((0, num_vert - 1))
            loc = ax.xaxis.get_major_locator()
            ind = loc().astype('int')
            ind = [i for i in ind if i >=0 and i < num_vert]

            nodes = list(range(0, graph.number_of_vertices()))

            ax.set_xticks(ind, [nodes[i] for i in ind])
        else:
            ax.set_xlim((0, num_vert - 1))
            loc = ax.xaxis.get_major_locator()
            ind = loc().astype('int')
            ind = [i for i in ind if i >=0 and i < num_vert]

            ax.set_xticks(ind)

    if min_prob is not None and max_prob is not None:
        plt.ylim((min_prob, max_prob*1.02))


def _plot_probability_distribution_on_graph(probabilities, ax, **kwargs):
    """
    Draw graph and illustrates the probabilities depending on the
    volatile parameters.

    See Also
    --------
    _update_nodes : sets volatile parameters
    :obj:`networkx.draw <networkx.drawing.nx_pylab.draw>`
    _configure_colorbar
    """

    # UpdateNodes may create kwargs['node_size']
    # min_node_size and max_node_size are not valid keys
    # for nx.draw kwargs
    _update_nodes(probabilities, kwargs.pop('min_node_size'),
                  kwargs.pop('max_node_size'), kwargs)

    vmin = kwargs.pop('min_prob')
    vmax = kwargs.pop('max_prob')
    nx.draw(kwargs.pop('graph'), ax=ax,
            node_size=kwargs.pop('node_size'),
            vmin=vmin, vmax=vmax, **kwargs)
    # Note: nx.draw_networkx_labels dramatically increases plotting time.
    # It is called by nx.draw
    kwargs['min_prob'] = vmin
    kwargs['max_prob'] = vmax

    # setting and drawing colorbar
    if 'cmap' in kwargs:
        _configure_colorbar(ax, kwargs)

    if __DEBUG__:
        global start
        end = time()
        start = end


def _configure_nodes(G, probabilities, kwargs):
    """
    Configure static attributes of nodes.

    Configures nodes attributes that will not change
    during multiple plots or an animation of a quantum walk.

    Parameters
    ----------
    kwargs: dict
        Reference to the dictionary **kwargs.
        Some entries may be added or altered.

    See Also
    --------
    _update_nodes
    """
    # setting colormap related attributes
    if 'cmap' in kwargs:
        if kwargs['cmap'] == 'default':
            kwargs['cmap'] = 'viridis'

    # setting node attributes
    if 'edgecolors' not in kwargs:
        kwargs['edgecolors'] = 'black'

    if 'linewidths' not in kwargs:
        kwargs['linewidths'] = 1

    if 'with_labels' not in kwargs:
        kwargs['with_labels'] = True

    if kwargs['with_labels'] and 'font_color' not in kwargs:
        kwargs['font_color'] = 'black'

    # calculates vertices positions.
    # needed to do beforehand in order to fix position for multiple steps
    # the user may choose any networkx graph_layout function
    # as long as only the graph is
    # the required parameter. Check
    # https://networkx.org/documentation/stable/reference/drawing.html#module-networkx.drawing.layout
    # For further customisation,
    # the user may call any networkx graph layout function
    # BEFORE calling plot_probability_distribution and
    # using its return as the 'pos' kwarg.
    if 'pos' not in kwargs:
        if 'graph_layout' in kwargs:
            func = kwargs.pop('graph_layout')
            kwargs['pos'] = func(G)
        else:
            kwargs['pos'] = nx.kamada_kawai_layout(G)


def _update_nodes(probabilities, min_node_size, max_node_size, kwargs):
    """
    Configure probability-related attributes of nodes.

    Configures attributes that may change depending on the probability
    at each (saved) step of the quantum walk.

    See Also
    --------
    _configure_nodes

    Notes
    -----
    The separation between UpdateNodes and ConfigureNodes optimizes
    plotting multiple images.
    """
    if 'cmap' in kwargs:
        kwargs['node_color'] = probabilities

    if 'node_size' not in kwargs:
        if min_node_size is None:
            min_node_size = 300
        if max_node_size is None:
            max_node_size = 3000

    if min_node_size is not None and max_node_size is not None:
        if ('rescale' in kwargs and kwargs.pop('rescale')):
            kwargs['min_prob'] = 0
            kwargs['max_prob'] = probabilities.max()

        # calculating size of each node acording to probability 
        # as a function f(x) = ax + b where b = min_size and
        # max_size = a*(max_prob-min_prob) + min_size
        a = ((max_node_size - min_node_size)
             / (kwargs['max_prob'] - kwargs['min_prob']))
        kwargs['node_size'] = list(map(
            lambda x: a*x + min_node_size, probabilities
        ))


def _configure_colorbar(ax, kwargs):
    """
    Add a colorbar in the figure besides the given ax

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        Ax on which the plot was drawn.
    kwargs : dict
        Dictionary containing the keys 'cmap', 'min_prob' and 'max_prob'.
        'min_prob' and 'max_prob' describe
        the inferior and superior limit for colorbar values, respectively.
        'cmap' describes a valid matplotlib colormap
    """
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    sm = plt.cm.ScalarMappable(
        cmap=kwargs['cmap'],
        norm=plt.Normalize(vmin=kwargs['min_prob'],
                           vmax=kwargs['max_prob'])
    )

    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='2.5%', pad=0.01)
    cbar = plt.colorbar(
        sm,
        ticks=np.linspace(kwargs['min_prob'], kwargs['max_prob'], num=5),
        cax=cax
    )

    cbar.ax.tick_params(labelsize=14, length=7)

def _default_plane_kwargs(kwargs):
    if not 'cmap' in kwargs:
        kwargs['cmap'] = 'default'

    if 'cmap' in kwargs:
        if kwargs['cmap'] == 'default':
            kwargs['cmap'] = 'viridis'

    if 'linewidth' not in kwargs:
        kwargs['linewidth'] = 0
    if 'antialiased' not in kwargs:
        kwargs['antialiased'] = False
    if 'cstride' not in kwargs:
        kwargs['cstride'] = 1
    if 'rstride' not in kwargs:
        kwargs['rstride'] = 1
    if 'alpha' not in kwargs:
        kwargs['alpha'] = 0.5

def _plot_probability_distribution_on_plane(
        probabilities, ax, labels=None, graph=None,
        min_prob=None, max_prob=None, dimensions=None, **kwargs
    ):
    """
    Plots probability distribution on the plane.
    """
    x_dim, y_dim = dimensions

    X = np.arange(0, x_dim, 1)
    Y = np.arange(0, y_dim, 1)
    X, Y = np.meshgrid(X, Y)
    Z = np.reshape(probabilities, (x_dim, y_dim))

    _default_plane_kwargs(kwargs)

    cmap = kwargs.pop('cmap')
    mappable = plt.cm.ScalarMappable(cmap=cmap)
    mappable.set_array(Z)
    if min_prob is not None and max_prob is not None:
        vmin = min_prob
        vmax = max_prob
    else: #rescale
        vmin = 0
        vmax = Z.max()
    mappable.set_clim(vmin, vmax)

    # division by 4 apparently normalize the colors
    ax.plot_surface(X, Y, Z, cmap=mappable.cmap,
                    vmin=vmin/4, vmax=vmax/4,
                    **kwargs)
    ax.set_zlim(vmin, vmax)
    kwargs['cmap'] = cmap # reinserts into kwargs

    cbar = plt.colorbar(mappable, shrink=0.4, aspect=20, pad=0.15)
    cbar.ax.tick_params(length=10, width=1, labelsize=16)

#########################################################################################

if __DEBUG__:
    start = time()
